import numpy as np
from hyvr.optimized import curve_interp
import matplotlib.pyplot as plt

def ferguson_curve(grid, h, k, ds, eps_factor, flow_angle, domain_length, xstart, ystart, width):
    """
    Simulate extruded parabola centrelines using the Ferguson (1976) disturbed meander model
    Implementation of AR2 autoregressive model
    http://onlinelibrary.wiley.com/doi/10.1002/esp.3290010403/full

    Parameters
    ----------
    grid : Grid object
    h : float
        Height (Ferguson model parameter)
    k : float
        Wave number (Ferguson model parameter)
    ds : float
        Curve distance for calculations (Ferguson model parameter)
    eps_factor : float
        Random background noise (Ferguson model parameter)
    flow_angle : float
        Angle of mean flow direction, in radians
    domain_length : float
        Length of the domain in mean flow direction
    xstart, ystart : float
        Starting coordinates of the channel centerline
    width : float
        Width of the channel

    Returns
    -------
    outputs : float matrix
        Simulated extruded parabola centerlines: storage array containing values for x coordinate, y
        coordinate, vx and vy
    """
    # Parameters
    ds += 1e-10
    ns = domain_length*2
    s = np.arange(0, ns, ds)

    # Centreline starting point
    xp = 0
    yp = 0

    # Calculate curve directions
    theta = ferguson_theta(s, eps_factor, k, h)

    # Interpolate curve direction over interval of interest
    s_interp, th_interp = curve_interp(s, theta, 0.1)

    # Storage array
    outputs = np.zeros((len(th_interp), 4))

    for th_idx, th_i in enumerate(th_interp):
        vx = ds*np.cos(th_i)
        vy = ds*np.sin(th_i)

        # Assign to storage array
        outputs[th_idx, 0] = xp       # x coordinate
        outputs[th_idx, 1] = yp       # y coordinate
        outputs[th_idx, 2] = vx       # vx
        outputs[th_idx, 3] = vy       # vy

        # go to next point
        xp += vx
        yp += vy

    # Rotate meanders into mean flow direction
    rot_angle = -np.mean(th_interp) + flow_angle
    rotMatrix = np.array([[np.cos(rot_angle), -np.sin(rot_angle)],
                          [np.sin(rot_angle),  np.cos(rot_angle)]])
    roro = np.dot(rotMatrix, outputs[:, 0:2].transpose())

    outputs[:, 2:] = np.dot(rotMatrix, outputs[:, 2:].transpose()).transpose()
    outputs[:, 0] = roro[0, :].transpose()
    outputs[:, 1] = roro[1, :].transpose()

    # move to channel start
    outputs[:, 0] += xstart
    outputs[:, 1] += ystart

    # plt.plot(outputs[:,0], outputs[:,1])
    # plt.axhline(y=grid.y0, xmin=grid.x0, xmax=grid.xmax, color='r')
    # plt.axhline(y=grid.ymax, xmin=grid.x0, xmax=grid.xmax, color='r')
    # plt.axvline(x=grid.x0, ymin=grid.y0, ymax=grid.ymax, color='r')
    # plt.axvline(x=grid.lx, ymin=grid.y0, ymax=grid.ymax, color='r')
    # plt.show()

    # Remove values before model domain
    indomain = np.logical_and(outputs[:, 0] >= grid.x0-width, outputs[:, 0] <= grid.xmax+width)
    outputs = outputs[indomain, :]

    # plt.plot(outputs[:,0], outputs[:,1])
    # plt.axhline(y=grid.y0, xmin=grid.x0, xmax=grid.xmax, color='r')
    # plt.axhline(y=grid.ymax, xmin=grid.x0, xmax=grid.xmax, color='r')
    # plt.axvline(x=grid.x0, ymin=grid.y0, ymax=grid.ymax, color='r')
    # plt.axvline(x=grid.lx, ymin=grid.y0, ymax=grid.ymax, color='r')
    # plt.show()

    return outputs


def ferguson_theta(s, eps_factor, k, h):
    """
    Calculate curve direction angle

    Parameters:
        s:				    Steps within generated curve distance
        eps_factor:		    Random background noise
        k:				    Wave number
        h:				    Height

    Returns:
        th_store *(array)* - Curve direction angle

    """
    # Storage arrays
    th_store = np.zeros(len(s))

    for idex, si in enumerate(s):
        if idex == 0:
            t1 = 0
            t2 = 0
            eps = 0
        elif idex == 1:
            t1 = th_store[idex-1]
            t2 = 0
            eps = np.random.normal()*eps_factor
        else:
            t1 = th_store[idex-1]
            t2 = th_store[idex-2]
            eps = np.random.normal(1)*eps_factor

        th_store[idex] = thetaAR2(t1, t2, k, h, eps)

    return th_store


def thetaAR2(t1, t2, k, h, eps):
    """
    Implementation of AR2 autoregressive model (Ferguson, 1976, Eq.15)
    http://onlinelibrary.wiley.com/doi/10.1002/esp.3290010403/full

    Parameters:
        t1: 	theta(i-1)
        t2: 	theta(i-2)
        k:		Wavenumber
        h:		Height
        eps:	Random background noise

    Returns:
        2nd-order autoregression (AR2)

    """
    b1 = 2*np.exp(-k*h)*np.cos(k*np.arcsin(h))
    b2 = -np.exp(-2*k*h)
    return eps + b1*t1 + b2*t2



import numpy as np
import hyvr.optimized as ho
import hyvr.utils as hu
from hyvr.classes.contact_surface import ContactSurface
from hyvr.classes.contact_surface_utils import *

def prob_choose(types, probs):
    # TODO: At the moment the given probabilities are only relative
    return np.random.choice(types, p=np.asarray(probs)/np.sum(probs))

class Stratum:

    def __init__(self, bottom_surface, top_surface, params, ae_types, grid):
        """
        Stratum constructor. This generates the stratum by generating the AEs
        within.

        Parameters
        ----------
        bottom_surface : ContactSurface object
            The bottom surface of the stratum
        top_surface : ContactSurface object
            The top surface of the stratum
        params : dict
            Dictionary with other necessary parameters
        ae_types : list of AEType objects
            The possible AE types within this stratum
        grid : Grid object
        """
        self.bottom_surface = bottom_surface
        self.top_surface = top_surface
        self.params = params
        self.name = params['strata']
        self.ae_types = ae_types
        self.ae_type_names = [ae_type.name for ae_type in ae_types]
        self.bg_facies = params['bg_facies']
        self.bg_azim = params['bg_azim']
        self.bg_dip = params['bg_dip']

        # generate AEs
        self.zmin = self.bottom_surface.zmin
        self.zmax = self.top_surface.zmax
        self.ae_zmins = []
        self.ae_zmaxs = []
        self.aes = []

        # first top is top of stratum
        ae_top_surface = self.top_surface
        # choose type of first AE
        curr_ae_type = prob_choose(self.ae_types, self.params['ae_prob'])
        znow = self.top_surface.zmean
        while znow > self.bottom_surface.zmean:

            num_type = self.ae_type_names.index(curr_ae_type.name)

            hu.print_to_stdout('Generating', curr_ae_type.name, 'from {:2.3f} m'.format(znow))

            # find height of AE
            mean_height = self.params['ae_z_mean'][num_type]
            # TODO: It might be nice to specify the height variance in the inifile
            height = np.random.normal(mean_height, mean_height*0.1)
            # potential avulsion
            if np.random.uniform() <= self.params['avul_prob'][0]:
                dz = -np.random.uniform(*self.params['avul'])
            else:
                dz = 0
            znow -= height + dz

            # create bottom contact surface
            # -----------------------------
            # the type of the bottom surface is determined by the type of the
            # AE below, if it is not the lowest AE
            next_ae_type = prob_choose(self.ae_types, self.params['ae_prob'])
            if znow <= self.bottom_surface.zmean:
                # limit top surface of uppermost AE to top surface of stratum
                znow = self.bottom_surface.zmean
                ae_bottom_surface = self.bottom_surface
            else:
                bottom_surface_params = next_ae_type.params['contact_model']
                bottom_surface_params['z'] = znow
                ae_bottom_surface = ContactSurface(grid, **bottom_surface_params)

            # if the bottom surface is above the top surface, we will assign
            # the (lower) top surface values instead
            use_lower_surface_value(ae_bottom_surface, ae_top_surface)

            # close to the bottom it might also be possible that the ae_bottom
            # is below the stratum bottom if the stratum bottom is not flat
            # in this case we will assign the (higher) bottom surface value instead
            use_higher_surface_value(ae_bottom_surface, self.bottom_surface)

            # generate element: this will place all the objects within an AE
            element = curr_ae_type.generate_realization(ae_bottom_surface, ae_top_surface, self, grid)
            self.aes.append(element)

            # if the minimum/maximum depth changed, update it
            ae_zmin = element.zmin
            if ae_zmin < self.zmin:
                self.zmin = ae_zmin
            self.ae_zmins.append(ae_zmin)
            ae_zmax = element.zmax
            if ae_zmax > self.zmax:
                self.zmax = ae_zmax
            self.ae_zmaxs.append(ae_zmax)

            # current bottom is top of next, next ae is new current ae
            ae_top_surface = ae_bottom_surface
            curr_ae_type = next_ae_type

        self.n_ae = len(self.aes)

    # def maybe_assign_facies_azim_dip(self, facies, angles, x, y, z, x_idx, y_idx, aei, oi, grid):
        # """
        # Assigns facies, azim and dip if the cell is inside this stratum or in
        # one of it's elements.

        # Since the objects inside the elements might reach into the domain
        # below, we can't just check the top and bottom surface but must check
        # every object.

        # Parameters
        # ----------
        # facies : np.ndarray[np.int32, dim=1]
            # Array of length one that hold the facies and will be altered by
            # this function. If this is -1 afterwards, the cell is not in the
            # stratum.
        # angles : np.ndarray[np.float, dim=1]
            # Array of length 2 that holds the azimuth and dip angles, will also
            # be altered.
        # x, y, z : float
            # x, y, and z positions of the cell in question
        # x_idx, y_idx : int
            # indices of x and y cells
        # aei : int
            # Index of where to start in the architectural element list. This is
            # typically the index of the last found element.
        # oi : int
            # Geometrical object index, similar to aei
        # grid : Grid object

        # Returns
        # -------
        # aei, oi : int
            # Updated indices
        # """
        # oi_orig = int(oi)
        # aei_orig = int(aei)
        # while aei < self.n_ae:
            # # if z > self.ae_zmaxs[aei]:
                # # if the current z is above the maximum z of the current AE, all other AEs will also
                # # be below the current cell, so we can stop searching for an AE
                # # This case should not happen
                # # break
            # if z < self.ae_zmins[aei]:
                # # if we're below the current AE, we can go search in the next one
                # oi = 0
                # aei += 1
                # continue
            # else:
                # self.aes[aei].maybe_assign_facies_azim_dip(
                    # facies, angles, x, y, z, x_idx, y_idx, oi, grid
                # )
                # if facies[0] != -1:
                    # # we found something
                    # return aei, oi
                # else:
                    # # if we didn't find something here, we check the next AE
                    # aei += 1
                    # oi = 0
                    # continue



import numpy as np
import hyvr.utils as hu

def generate_trough_positions(bottom_surface, top_surface, type_params, grid):
    # TODO:
    # Idea: Generating the objects on planes between the maximum of the
    # bottom surface and the top surface is not exactly perfect imho.
    # Instead, we could maybe calculate the volume of the AE and use
    # this for generating the number of troughs, e.g.:
    #
    #     trough_density = #troughs/area = (#throughs * height) / volume
    #
    # The height should be the 'agg' parameter. Therefore, the number
    # of troughs could be generated according to a poisson
    # distribution:
    #
    #     n_troughs = np.random.poisson(lam=trough_density * volume/agg)
    #
    # For migrating troughs, we could then first assign the first half
    # in the "upper part" (what should this be?) and the second half as
    # migrated troughs based on the previous ones.

    # TODO:
    # Also randomly place objects a bit outside the domain if the domain is not
    # periodic. Therefore we need a and b

    # Randomly generated trough positions
    bottom_z = bottom_surface.zmax + type_params['depth'] * type_params['buffer']
    top_z = top_surface.zmean
    # go from top to bottom
    planes_z = np.arange(top_z, bottom_z, -type_params['agg'])

    # construct list of random trough centers
    te_xyz = []
    for i, z in enumerate(planes_z):
        if i == 0 or type_params['migrate'] is None:
            # randomly draw positions according to spatial uniform distribution
            lam = type_params['trough_density'] * grid.lx * grid.ly
            n_troughs = np.random.poisson(lam=lam)
            for k in range(n_troughs):
                xc = np.random.uniform(grid.x0, grid.xmax)
                yc = np.random.uniform(grid.y0, grid.ymax)
                te_xyz.append([xc, yc, z])

        else:   # migrate previous troughs
            # n_troughs stays the same
            for k in range(n_troughs):
                above_idx = k + (i-1)*n_troughs
                xc = te_xyz[above_idx][0] + np.random.normal(type_params['migrate'][0],
                                                             type_params['migrate'][1])
                yc = te_xyz[above_idx][1] + np.random.normal(type_params['migrate'][2],
                                                             type_params['migrate'][3])
                te_xyz.append([xc, yc, z])

    return te_xyz


def rand_trough_params(type_params, z, grid):
    if type_params['size_ztrend'] is not None:
        delta_ztrend = type_params['size_ztrend'][1] - type_params['size_ztrend'][0]
        zfactor = type_params['size_ztrend'][0] +\
            (z - grid.z0)*(delta_ztrend)/grid.lz
    else:
        zfactor = 1

    params = {}
    params['a'] = type_params['length'] * zfactor / 2
    params['b'] = type_params['width'] * zfactor / 2
    params['c'] = type_params['depth'] * zfactor
    params['alpha'] = np.random.uniform(*type_params['paleoflow'])

    return params



"""
This module contains the classes Section and Option to simplify parsing and
validating ini-files.

:Author: Samuel Scherrer
"""

import sys
from copy import deepcopy

__all__ = ["Section", "Option", "MissingSectionError", "MissingOptionError", "ShapeError", "assert_exists"]


class Section():

    def __init__(self, name, options):
        """
        Parameters
        ----------
        name : str
            Name of the section
        options : list
            List of Options
        """
        self.name = name
        self.options = deepcopy(options)
        self.optionnames = [o.name for o in options]

    def parse(self, section_dict):
        """
        Parses and validates options of the section given a dictionary
        containing key-value pairs of the options. If options are not present,
        default values might be set.

        Parameters
        ----------
        section_dict : dictionary
            Dictionary of section values. This can for example be obtained
            using ``configparser``::

                p = configparser.ConfigParser()
                p.read(filename)
                section_dict = dict(p[section_name])

        Returns
        -------
        section_dict : dictionary
            The same dictionary with parsed and validated values
        """
        self.dict = section_dict
        for option in section_dict:
            if option not in self.optionnames:
                print("Warning: Unknown option: {:s} in section {:s}".format(
                    option, self.name), file=sys.stderr
                )
        for option, name in zip(self.options, self.optionnames):
            self.dict[name] = option.parse(self)
        return self.dict

    def __repr__(self):
        repr =  "Section(name={:s},options=".format(self.name)
        for name in self.optionnames:
            repr += name + ','
        repr += ')'
        return repr


class Option():
    """
    Parameters
    ----------
    name : string
        name of the option
    dtype : type
        type of the value of the option, e.g. float, str, int, or list.
        If dtype is list, every entry of the list must have the same type.
    optional : bool, optional (default: False)
        whether the option is optional or not
    default : optional, (default: None)
        if optional=True, this default value will be used if this option is not
        given.
    shape : int or string or list/tuple of ints and/or strings, optional (only
        required for lists, default: None)
        If dtype is ``list``, this is the shape of the list.
        There are several possibilities how to use this option:

        * if ``shape=n`` where n is a nonnegative integer, the value must be
          a list with length ``n``.
        * if ``shape=-1`` the value can have an arbitrary shape.
        * if ``shape="option1"``, the value of this option must have the same
          shape as "option1". This is especially useful if the shape of
          "option1" is set to -1.
        * if ``shape=[2, 3]``, the value must be a list of lists, where the
          outermost list has length 2 and the inner lists all have length 3.
          This also works for more than 2 dimensions.
        * if ``shape=[2, -1, 3]``, the value must be a list of lists of lists.
          The outermost list must again have length 2, the innermost lists must
          have length 3, and the lists at the intermediate level can have any
          length (even different lengths).
        * if ``shape=[2, "option1", 3]``, the values must again be a list of
          lists similar to above, but now the lists at the intermediate level
          must have the same length as "option1".
        * if ``shape=[2, [1, 2], [[3], [3, 3]]]``, the value must be a list
          of lists of lists. The outermost list must again have length 2. The
          two lists it contains have length 1 and length 2. The innermost lists
          all must have length 3.

        It's also possible to only give the innermost value(s), e.g. for a list
        with ``shape=[2, 3]`` only the value ``18``. This will then be expanded
        to ``[[18, 18, 18], [18, 18, 18]]``. Similarly, ``[18, 19, 20]`` would
        be expanded to ``[[18, 19, 20], [18, 19, 20]]``.
        This expansion obviously doesn't work if ``shape=-1``.
        If ``shape=[2, -1, 3]``, expansion is possible if the given value is
        e.g. ``[[1, 2, 3], [1, 2, 3], [1, 2, 3]]``, but not if only ``[1, 2, 3]``
        is given, since the length of the second dimension must be determined
        from the given value.
    datatype: int, float or string, optional (only required for lists, default: None)
        Type of the innermost elements of the option in case the option is a list.
    validation_func: function of one argument that returns a boolean, optional.
        Returns true if the value (for lists or lists of lists this applies to
        all values) is valid.
    alternatives: list of strings
        List of alternative names for this option. If the option is not found
        in the given dictionary while parsing, these alternatives are used if
        they exist. If an alternative is found (searching happens in the
        supplied order), the option is stored under it's standard name.
    """

    def __init__(self, name, dtype, optional=False, default=None, shape=-1,
                 datatype=None, validation_func=lambda x: True, alternatives=[]):
        self.name = name
        self.dtype = dtype
        self.optional = optional
        self.default = default
        self.shape = shape
        self.datatype = datatype
        self.validation_func = validation_func
        if not isinstance(alternatives, list):
            alternatives = [alternatives]
        self.alternatives = alternatives

        # make sure we have shape and datatype for lists
        if self.dtype == list:
            if self.shape is None:
                raise ValueError('Option ' + self.name + ' has type list, but no shape was given.')
            if self.datatype is None:
                raise ValueError('Option ' + self.name + ' has type list, but no datatype for its elements was given.')

    def __repr__(self):
        return "Option(name={:s}, dtype={:s}, optional={:s}, default={:s}, shape={:s}, datatype={:})".format(
            self.name, str(self.dtype), str(self.optional), str(self.default), str(self.shape), str(self.datatype)
        )

    def parse(self, section):
        """
        Parses the option based on it's attributes.
        """
        # try to find alternatives if they exist
        alternatives = deepcopy(self.alternatives)
        while len(alternatives) != 0 and self.name not in section.dict:
            other_name = alternatives.pop(0)
            if other_name in section.dict:
                section.dict[self.name] = section.dict[other_name]
                del section.dict[other_name]
                break
        if not self.optional:
            assert_exists(self.name, section.dict, section.name)
        if self.name not in section.dict:
            return self.default
        else:
            if self.dtype != list:
                if self.dtype == bool:
                    # this is necessary since ``bool("False")`` returns ``True``.
                    value = parse_bool(section, self.name)
                else:
                    value = self.dtype(section.dict[self.name])
                if not self.validation_func(value):
                    raise ValueError('Invalid input for option ' + self.name +
                                    ' in section ' + section.name)
                return value
            else:

                value = parse_list(section.dict[self.name], self.datatype)

                # value validation
                if not all_true(self.validation_func, value):
                    raise ValueError('Invalid input for option ' + self.name +
                                    ' in section ' + section.name)

                shape = deepcopy(self.shape)

                # now we need to get the correct shape
                if shape == -1:
                    # we don't care for the shape of this
                    if not isinstance(value, list):
                        value = [value]
                    return value

                if isinstance(shape, str):
                    # in this case we simply use the shape of the option with this name
                    if shape not in section.dict:
                        raise ValueError(self.name + ' in ' + section.name + ' has an invalid ' +\
                                         'shape because the options whose shape it should have ' +\
                                         'does not exist. Check your option definitions!')
                        raise ShapeError(self.name, section.name)
                    shape = get_shape(section.dict[shape])
                if isinstance(shape, int):
                    shape = [shape]
                # shape is now a list, but it might still contain strings
                for i in range(len(shape)):
                    if isinstance(shape[i], str):
                        shape[i] = len(section.dict[shape[i]])



                # shape is now either a 'flat' shape, i.e. something like [2, 3, 2],
                # or an expanded shape, e.g. [2, [3, 3], [[2, 2, 2],[2, 2, 2]]]
                # if it's flat, it might contain dimensions with -1 that cannot be
                # autoexpanded. We first need to determine the shape of this dimension.
                if is_flat(shape):
                    real_shape = get_shape(value)
                    if isinstance(real_shape, (list, tuple)):
                        # if it's just a single number we can expand it
                        # Here I'm trying to find the flat shape of the value that was
                        # given in the configuration file.
                        flat_shape_value = try_flattening_shape(real_shape)
                        # It might happen that we cannot flatten the shape, in this
                        # case there are negative values remaining in flat_shape_value.
                        # If there are, this means that there is a dimension
                        # containing lists of different lengths.
                        # In any case I will try to replace any -1 in ``shape``
                        # with the value in ``flat_shape_value``.
                        shape = get_positive_shape(shape, flat_shape_value)
                        # Now we do a test for equality of the asserted shape and
                        # the shape of the value found in the config file. Keep in
                        # mind that there might be -1 values left.
                        if flat_shape_value != shape[-len(flat_shape_value):]:
                            raise ShapeError(self.name, section.name)
                        # If there are -1's left we must ensure that the "depth" of
                        # the given value, i.e. the number of dimensions, is higher
                        # than the ``number of dimensions after the value preceding
                        # the first -1`` + 1 .
                        if any(map(lambda x: x == -1, shape)):
                            depth = numdim(value)
                            mindepth = len(shape) - shape.index(-1) + 1
                            if depth < mindepth:
                                raise ValueError('Option ' + self.name + ' in section ' +
                                                section.name + ' can not be expanded!')
                        shape = expand_shape(shape)

                # Now we have an expanded shape, so only two tasks remain:
                # * auto-expansion
                # * shape validation
                value = expand_to_shape(shape, value)
                if not compare_shapes(shape, get_shape(value)):
                    raise ShapeError(self.name, section.name)
                return value



def numdim(l):
    """
    Returns number or dimensions of the list, assuming it has the same depth
    everywhere.
    """
    if not isinstance(l, (list, tuple)):
        return 0
    if not isinstance(l[-1], (list, tuple)):
        return 1
    else:
        return 1 + numdim(l[-1])


def _get_shape(l):
    depth = numdim(l)
    if depth == 0:
        return -1
    if depth == 1:
        return len(l)
    else:
        return [_get_shape(s) for s in l]

def get_shape(l):
    """
    Returns the expanded shape of a nested list, e.g.:

    >>> get_shape(42)
    1
    >>> get_shape([1, 2])
    [2]
    >>> get_shape([[1, 2], [3, 4, 5]])
    [2, [2, 3]]
    >>> get_shape([ [[1, 2], [3, 4]] , [[5, 6, 7], [8]] ])
    [2, [2, 2], [[2, 2], [3, 1]]]
    """
    s = _get_shape(l)
    if s == -1:
        return 1
    shape = [s]
    for i in range(numdim(s)):
        s = _get_shape(s)
        shape.append(s)
    return shape[::-1]

def is_flat(shape):
    return all(map(lambda x: numdim(x) == 0, shape))

def try_flattening_shape(shape):
    depth = numdim(shape)
    flat = [shape[0]]
    for i in range(1, depth):
        flat_list = unroll(shape[i])
        if all_true(lambda x: x == flat_list[0], flat_list):
            flat.append(flat_list[0])
        else:
            flat.append(-1)
    return flat

def get_positive_shape(shape_in_option, shape_in_file):
    """
    Returns a flat shape without -1, replacing the first part with the given
    shape in the option if it's only partially given in the file.
    """
    shape = deepcopy(shape_in_option)
    for i in range(len(shape_in_file)):
        if shape[len(shape)-1-i] == -1:
            shape[len(shape)-1-i] = shape_in_file[len(shape_in_file)-1-i]
    return shape




def unroll(l):
    if numdim(l) == 0:
        return [l]
    if numdim(l) == 1:
        return l
    else:
        unrolled = []
        for li in l:
            unrolled += unroll(li)
        return unrolled


def expand_shape(shape):
    """
    Expands a flat shape to an expanded shape

    >>> expand_shape([2, 3])
    [2, [3, 3]]
    >>> expand_shape([2, 3, 4])
    [2, [3, 3], [[4, 4, 4], [4, 4, 4]]]
    """
    expanded = [shape[0]]
    for i in range(1, len(shape)):
        next = [shape[i]] * shape[i-1]
        for j in range(1, i):
            next = [next] * shape[j]
        expanded.append(next)
    return expanded

def expand_to_shape(shape, value):
    depth = numdim(value)
    if depth != len(shape):
        value = _expand_to_shape(shape[-1-depth], value)
    return value

def _expand_to_shape(shape, value):
    if isinstance(shape, int):
        value = [value] * shape
        return value
    else:
        return [_expand_to_shape(s, value) for s in shape]

def compare_shapes(s1, s2):
    d1 = numdim(s1)
    d2 = numdim(s2)
    if d1 != d2:
        return False
    else:
        l1 = unroll(s1)
        l2 = unroll(s2)
        if len(l1) != len(l2):
            return False
        else:
            for i in range(len(l1)):
                if l1[i] != l2[i] and l1[i] != -1 and l2[i] != -1:
                    return False
            return True



def assert_exists(option, dictionary, section_name):
    if option not in dictionary:
        raise MissingOptionError(option, section_name)

def all_true(func, l):
    if not isinstance(l, (list,tuple)):
        return func(l)
    else:
        return all([all_true(func, li) for li in l])


def parse_list(string, dtype):
    """
    Parses string of form ``'[as, df]'`` to list of type ``dtype``, e.g. for
    ``dtype=str`` it returns `['as', 'df']``. This works also for a list of lists.

    Parameters
    ----------
    string : string
        Has form ``'[some, entries]``. All entries in the list must have the same type.
    dtype : type converter
        Converter to the type the entries should have, e.g. ``str`` or ``int``

    Returns
    -------
    list of elements of type ``dtype`` or list of lists of elements of type
    ``dtype``. Returns a single value if only a single value is given.
    """
    # l =  string.replace('[', '').replace(']', '').replace(' ', '').split(',')
    s = string.replace(' ', '') # remove all spaces first
    if s[0] == '[': # it's not only a single item
        s = s[1:-1] # remove [ and ] from start and end only
    else: # it's just a single item
        return dtype(s)
    if s[0] == '[': # it's a list of lists
        splitted = s.split('],')
        for i in range(len(splitted)-1):
            splitted[i] += ']' # splitting removed the closing bracket from all but the last item
        l = list(map(lambda x: parse_list(x, dtype), splitted))
    else:
        splitted = s.split(',')
        l = list(map(dtype, splitted))
    return l

def parse_bool(section, optionname):
    """
    Parses a string option as bool. Possible options are "True"/"False",
    "yes"/"no", "1"/"0".
    """
    string = section.dict[optionname]
    if string.lower() == "true" or string.lower() == "yes":
        return True
    elif string.lower() == "false" or string.lower() == "no":
        return False
    elif string.isdigit():
        return bool(int(string))
    else:
        raise ValueError("Option " + optionname + " in section " + section.name
                         + " is not a valid boolean!")



class ShapeError(Exception):
    def __init__(self, optionname, sectionname):
        message = "Option " + optionname + " in section " + sectionname + " has wrong shape!"
        super().__init__(message)

class MissingOptionError(Exception):
    def __init__(self, optionname, sectionname):
        message = "Option " + optionname + " in section " + sectionname + " is missing!"
        super().__init__(message)

class MissingSectionError(Exception):
    def __init__(self, sectionname):
        message = "Section " + sectionname + " is missing!"
        super().__init__(message)




import numpy as np
import hyvr.utils as hu

def gen_extpar(channel_type, model, ae, count, ani=True):
    """
    Generate extruded parabola geometries:

    * Flow regime is assumed to be reasonably constant so the major geometry of the extruded
      parabolas doesn't change so much
    * 'Migration' of the extruded parabolas according to a shift vector


    Parameters
    ----------
    extpar_type : ExtrudedParabolaType
    model : Model object
    ae : AERealization object
    ae : list
        AE parameters
        TODO: use AERealization instead
    count : int
        Material number and/or identifier
    ani : bool, optional (default: True)
        Whether anisotropy should be generated
        
    Returns
    -------
    count : int
        Material number and/or identifier
    """
    # This is just for convenience
    ae_arr = model.data['ae']
    ha_arr = model.data['ha']
    hat_arr = model.data['hat']
    fac = model.data['fac']
    azim = model.data['azim']
    dip = model.data['dip']

    x2 = model.X[:,:,0]
    y2 = model.Y[:,:,0]
    # TODO: replace z3 with model.Z
    _, _, z3 = np.meshgrid(range(0, model.nx), range(0, model.ny), range(0, model.nz), indexing='ij')          # 2-D grid

    # start location
    total_extpar = int(channel_type.channel_no)
    if model.display:
        # Place troughs in domain centre for display features
        xstart = (model.x0 + model.lx)/2 * np.ones(total_extpar)
        ystart = np.random.uniform(model.y0, model.y0+model.ly, total_extpar)
    else:
        # Randomly place curve starting points
        # TODO: this was the original setting for xstart, why was this hard-coded?
        # xstart = np.random.uniform(-10, 0, total_extpar)
        xstart = np.random.uniform(model.x0, model.xmax, total_extpar)
        ystart = np.random.uniform(model.y0, model.ymax, total_extpar)

    # loop over extruded parabola top depths
    ch_bot = ae.z_bottom
    ch_bot += channel_type.depth * channel_type.buffer

    ch_top = ae.z_top
    ch_dz = channel_type.agg
    ch_top += ch_dz

    for znow in np.arange(ch_bot, ch_top, ch_dz):
        hu.print_to_stdout('z = {:2.4f}m'.format(znow))
        # Assign linear trend to extruded parabola sizes
        if channel_type.geo_ztrend is not None:
            zfactor = np.interp(znow, [model.z0, model.z0 + model.lz], [channel_type.geo_ztrend[0], channel_type.geo_ztrend[1]])
        else:
            zfactor = 1
        z_ch_width = channel_type.width * zfactor
        z_ch_depth = channel_type.depth * zfactor

        if channel_type.channel_start is not None:
            cstart = channel_type.channel_start
        else:
            cstart = []
        # Loop over total extruded parabolas per system
        for chan in range(0, total_extpar):
            """ Loop over multiple extruded parabolas at 'timestep' """
            aha = ferguson_curve(model, channel_type.h, channel_type.k,  channel_type.ds,
                                 channel_type.eps_factor,
                                 disp=model.display,
                                 ch_start=cstart)

            """ Get flow direction in extruded parabolas for azimuth """
            # For periodicity shift trajectories into model unit cell
            if model.periodic:
                aha[aha[:, 1] < yvec[0], 1] += model.ly
                aha[aha[:, 1] > yvec[-1], 1] -= model.ly

            # initialize 2-D distance matrix
            D = 1e20 * np.ones_like(x2)

            # initialize sum of inverse-distance weights
            sumW = np.zeros(np.shape(x2), dtype=float)
            W = np.zeros(np.shape(x2), dtype=float)

            # initialize velocity orientation at this level
            vx_znow = np.zeros(np.shape(x2), dtype=float)
            vy_znow = np.zeros(np.shape(x2), dtype=float)

            # loop over all points of trajectory
            for ii in range(0, len(aha)):
                # distance to current point
                R = np.sqrt((x2 - aha[ii][0]) ** 2 + (y2 - aha[ii][1]) ** 2)

                # smallest distance of entire grid to all points so far
                D[R < D] = R[R < D]

                # inverse-distance weight for velocity interpolation
                W[:] = 1e-20
                W[R < z_ch_width / 2] = 1 / (R[R < z_ch_width / 2] + 1e-20)

                # velocity interpolation in 2-D
                vx_znow += aha[ii][2] * W
                vy_znow += aha[ii][3] * W
                sumW += W

            vx_znow /= sumW
            vy_znow /= sumW

            # Assign facies sets with dip values
            if sum(channel_type.dip) > 0:
                do, fd, dv, av = hu.dip_sets(model, channel_type, znow, curve=[aha[:, 0], aha[:, 1], vx_znow, vy_znow])
            else:
                do = np.ones((model.nx, model.ny, model.nz)) + count
                fd = np.ones((model.nx, model.ny, model.nz), dtype=int) * int(np.random.choice(channel_type.facies))
                dv = 0.0
                av = np.zeros((model.nx, model.ny, model.nz))

            """ Copy results into 3-D field """
            # Iterate over all nodes below current top elevation
            znow_index = int(np.around((znow - model.z0)/model.dz)) # TODO: fix around stuff
            zbelow_index = int(np.around((znow - z_ch_depth - model.z0)/model.dz))
            d_range = np.arange(max(0, zbelow_index), min(model.nz, znow_index))      # Depth range
            if len(d_range) > 0:        # Only compute if extruded parabola depth range is finite

                # Get mask arrays for each condition
                in_extpar = D[:, :, None]**2 <= z_ch_width**2 / 4 - ((znow_index - z3) * model.dz * z_ch_width / (z_ch_depth*2)) ** 2     # is grid cell in extruded parabola
                finite_v = ~np.isnan(vx_znow)            # Only assign if velocity is finite
                below_top = ae_arr <= ae.num          # Don't assign values to locations higher than top contact surface
                chan_mask = in_extpar * finite_v[:, :, None] * below_top
                if znow_index <= z3[:, :, -1].max():
                    # Set mask above top of extruded parabola to False
                    chan_mask[:, :, znow_index:-1] = False

                # Assign properties
                fac[chan_mask] = fd[chan_mask]
                ha_arr[chan_mask] = count
                hat_arr[chan_mask] = channel_type.ae_id
                ae_arr[chan_mask] = ae.num

                if channel_type.lag is not None:
                    in_lag = (znow - z_ch_depth + float(channel_type.lag[0])) > z3 * model.dz   # Is grid cell in extruded parabola
                    fac[np.logical_and(in_extpar, in_lag)] = int(channel_type.lag[1])

                if ani:
                    # calcuate azimuth, to 1 degree
                    azim2d = np.round((np.arctan2(vx_znow, vy_znow) - np.pi/2) * 180/np.pi)
                    azim3d = azim2d[:, :, None] * np.ones((model.nx, model.ny, model.nz))
                    azim[chan_mask] = azim3d[chan_mask]
                    dip[chan_mask] = dv

                count += 1

        # Shift starting values with migration vector from parameter file
        if channel_type.mig: # TODO: fix this
            xstart += np.random.uniform(-channel_type.mig[0], channel_type.mig[0])
            ystart += np.random.uniform(-channel_type.mig[1], channel_type.mig[1])

    return count


def ferguson_curve(model, h, k, ds, eps_factor, dist=0, disp=False, ch_start=[]):
    """
    Simulate extruded parabola centrelines using the Ferguson (1976) disturbed meander model
    Implementation of AR2 autoregressive model
    http://onlinelibrary.wiley.com/doi/10.1002/esp.3290010403/full

    Parameters
    ----------
    model : Model object
    h : float
        Height
    k : float
        Wave number
    ds : float
        Curve distance for calculations
    eps_factor : float
        Random background noise
    dist : float, optional (default: 0)
        Distance to generate curves - defaults to model.lx (if dist=0)
    disp : bool, optional (default: False)
        Creating display extruded parabola - extruded parabola begins at (0,0)
    ch_start : tuple of floats
        Starting location of channel (x,y coordinates)

    Returns
    -------
    outputs : numpy array 
        Simulated extruded parabola centerlines: storage array containing values for x coordinate, y
        coordinate, vx and vy
    """
    # Parameters
    ds += 1e-10
    if dist > 0:
        ns = dist
    else:
        ns = model.lx*100
    s = np.arange(0, ns, ds)

    # Centreline starting point
    xp = 0
    yp = 0

    # Calculate curve directions
    theta = ferguson_theta(s, eps_factor, k, h)

    # Interpolate curve direction over interval of interest
    s_interp, th_interp = curve_interp(s, theta, 0.1)

    # Storage array
    outputs = np.zeros((len(th_interp), 4))

    for th_idx, th_i in enumerate(th_interp):
        vx = ds*np.cos(th_i)
        vy = ds*np.sin(th_i)
        xp += vx
        yp += vy

        # Assign to storage array
        outputs[th_idx, 0] = xp       # x coordinate
        outputs[th_idx, 1] = yp       # y coordinate
        outputs[th_idx, 2] = vx       # vx
        outputs[th_idx, 3] = vy       # vy

    # Rotate meanders into mean flow direction
    mean_th = -np.mean(th_interp)
    rotMatrix = np.array([[np.cos(mean_th), -np.sin(mean_th)],
                          [np.sin(mean_th),  np.cos(mean_th)]])
    roro = np.dot(rotMatrix, outputs[:, 0:2].transpose())

    outputs[:, 2:] = np.dot(rotMatrix, outputs[:, 2:].transpose()).transpose()

    # Move starting location in x-direction
    outputs[:, 0] = roro[0, :].transpose() - np.random.uniform(model.lx/5, model.lx/10)
    outputs[:, 1] = roro[1, :].transpose()

    # Remove values before model domain
    if dist > 0:
        indomain = outputs[:, 0] >= model.x0
    else:
        indomain = np.logical_and(outputs[:, 0] >= model.x0, outputs[:, 0] <= model.lx)
    outputs = outputs[indomain, :]

    if len(ch_start) > 0:
        outputs[:, 0] += ch_start[0]
        outputs[:, 1] += ch_start[1]

    # Make sure streamlines begin within domain with respect to y
    yout = outputs[0, 1] > model.ly/4 or outputs[0, 1] > model.ly/4
    if disp is True:
        starty = outputs[0, 1]
    elif yout is True:
        starty = np.random.uniform(-model.ly/4, model.ly/4)
    else:
        starty = 0
    outputs[:, 1] = outputs[:, 1] - starty

    return outputs


def ferguson_theta(s, eps_factor, k, h):
    """
    Calculate curve direction angle

    Parameters:
        s:				    Steps within generated curve distance
        eps_factor:		    Random background noise
        k:				    Wave number
        h:				    Height

    Returns:
        th_store *(array)* - Curve direction angle

    """
    # Storage arrays
    th_store = np.zeros(len(s))

    for idex, si in enumerate(s):
        if idex == 0:
            t1 = 0
            t2 = 0
            eps = 0
        elif idex == 1:
            t1 = th_store[idex-1]
            t2 = 0
            eps = np.random.normal()*eps_factor
        else:
            t1 = th_store[idex-1]
            t2 = th_store[idex-2]
            eps = np.random.normal(1)*eps_factor

        th_store[idex] = thetaAR2(t1, t2, k, h, eps)

    return th_store


def thetaAR2(t1, t2, k, h, eps):
    """
    Implementation of AR2 autoregressive model (Ferguson, 1976, Eq.15)
    http://onlinelibrary.wiley.com/doi/10.1002/esp.3290010403/full

    Parameters:
        t1: 	theta(i-1)
        t2: 	theta(i-2)
        k:		Wavenumber
        h:		Height
        eps:	Random background noise

    Returns:
        2nd-order autoregression (AR2)

    """
    b1 = 2*np.exp(-k*h)*np.cos(k*np.arcsin(h))
    b2 = -np.exp(-2*k*h)
    return eps + b1*t1 + b2*t2


def curve_interp(xc, yc, spacing):
    """
    Interpolate evenly spaced points along a curve. This code is based on code in an answer posted by 'Unutbu' on
    http://stackoverflow.com/questions/19117660/how-to-generate-equispaced-interpolating-values (retrieved 17/04/2017)

    Parameters:
        xc:			x coordinates of curve
        yc:			y coordinates of curve
        spacing:	Spacing between points

    Returns:
        - xn - x coordinates of interpolated points
        - yn - y coordinates of interpolated points

    """

    # TODO: what does this function do exactly? Can we use np.interp instead?
    N = len(xc)
    xc_in = np.array(xc)
    yc_in = np.array(yc)

    t = np.arange(xc[0], N, spacing * 0.1)
    xc = np.interp(t, np.arange(N), xc)
    yc = np.interp(t, np.arange(N), yc)
    tol = spacing
    ic, idx = 0, [0]
    while ic < N:
        total_dist = 0
        for j in range(ic+1, N):
            total_dist += np.sqrt((xc[j] - xc[j-1]) ** 2 + (yc[j] - yc[j-1]) ** 2)
            if total_dist > tol:
                idx.append(j)
                break
        ic = j + 1

    xn = xc[idx]
    yn = yc[idx]
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()
    ax.plot(xc_in, yc_in, 'rx-')
    ax.scatter(xn, yn)
    # ax.set_aspect('equal')
    plt.show()

    return xn, yn



"""
This module contains all options for the ini-files. If you want to add an option, add it here and potentially do some post-processing in parameters.

:Authors: Samuel Scherrer, Jeremy P. Bennett
"""


from hyvr.input.option_parsing import Option

options = {}

# Run options
# ===========
options['run'] = [
    Option('runname', str, optional=True, default=None),
    Option('modeldir', str, optional=True, default=None),
    Option('numsim', int, optional=True, default=1),
    Option('anisotropy', bool, optional=True, default=True,
           alternatives='flag_anisotropy'),
    Option('het', bool, optional=True, default=True,
           alternatives='flag_het'),
    Option('dataoutputs', list, optional=True, shape=[-1], default=[], datatype=str,
           validation_func=lambda x: x in ['vtr', 'py', 'mat', 'npz', 'h5'],
           alternatives='l_dataoutputs'),
    Option('modeloutputs', list, optional=True, shape=[-1], default=[], datatype=str,
           validation_func=lambda x:  x in ['mf', 'mf6', 'hgs'],
           alternatives='l_modeloutputs'),
    Option('flag_ow', bool, optional=True, default=False),
]


# Model options
# =============
options['model'] = [
    Option('x0', float, optional=True, default=0.0),
    Option('y0', float, optional=True, default=0.0),
    Option('z0', float, optional=True, default=0.0),
    Option('dx', float, optional=False),
    Option('dy', float, optional=True), # use value of 'dx' if 'dy' is missing
    Option('dz', float, optional=True),
    Option('lx', float, optional=False),
    Option('ly', float, optional=False),
    Option('lz', float, optional=False),
    Option('periodic', bool, optional=True, default=False,
           alternatives='flag_periodic'),
    Option('display', bool, optional=True, default=False,
           alternatives='flag_display'),
    Option('hetlev', str, optional=False,
           validation_func=lambda x: x in ['ae', 'facies', 'internal']),
]


# Strata options
# ==============
options['strata'] = [
    Option('ssm', list, optional=False, shape=[-1], datatype=str,
           alternatives=['l_ssm', 'l_seq']),
    Option('ssm_top', list, optional=False, shape='ssm', datatype=float,
           alternatives='r_ssm_top'),
    Option('ssm_contact_model', list, optional=True, shape=['ssm', 3], datatype=float,
           alternatives='ll_ssm_contact_model'),
    Option('ssm_contact', str, optional=True, default='flat',
           validation_func=lambda x: x in ['flat', 'random', 'user']),
    Option('ae_table', str, optional=True, default=None),
    Option('save_aelu', bool, optional=True, default=False),
    Option('ae', list, optional=False, shape=[-1], datatype=str,
           alternatives='l_ae'),
    Option('ssm_ae', list, optional=False, shape=['ssm', -1], datatype=str,
           alternatives='ll_ssm_ae'),
    Option('ae_prob', list, optional=False, shape='ssm_ae', datatype=float,
           alternatives='ll_ae_prob'),
    Option('ae_z_mean', list, optional=False, shape='ssm_ae', datatype=float,
           alternatives='ll_ae_z_mean'),
    Option('avul_prob', list, optional=False, shape=['ssm', 1], datatype=float,
           alternatives='ll_avul_prob'),
    Option('avul', list, optional=False, shape=['ssm', 2], datatype=float,
           alternatives='ll_avul'),
    Option('bg', list, optional=False, shape=3, datatype=float),
]


# Element options
# ===============
element_options = [
    Option('geometry', str, optional=False,
           validation_func=lambda x: x in ['trunc_ellip', 'ext_par', 'sheet']),
    Option('structure', str, optional=False),
    Option('contact', str, optional=False,
           validation_func=lambda x: x in ['flat', 'random']),
    Option('contact_model', list, optional=True, shape=[3], datatype=float,
           alternatives='r_contact_model'),
    Option('facies', list, optional=False, shape=[-1], datatype=int,
           alternatives='l_facies'),
    Option('altfacies', list, optional=True, shape=['facies', -1], datatype=int,
           alternatives='ll_altfacies'),
    Option('geo_ztrend', list, optional=True, shape=2, datatype=float,
           alternatives='r_geo_ztrend'),
    Option('k_ztrend', list, optional=True, shape=2, datatype=float,
           alternatives='r_k_ztrend'),
    Option('k_xtrend', list, optional=True, shape=2, datatype=float,
           alternatives='r_k_xtrend'),
    Option('n_xtrend', list, optional=True, shape=2, datatype=float,
           alternatives='r_n_xtrend'),
    Option('n_ztrend', list, optional=True, shape=2, datatype=float,
           alternatives='r_n_ztrend'),
    Option('bg', list, optional=True, shape=3, datatype=float)
]

erosive_element_options = [
    Option('agg', float, optional=False),
    Option('buffer', float, optional=True, default=0.0),
    Option('dipset_d', float, optional=True),
    Option('migrate', list, optional=True, shape=4, datatype=float,
           alternatives='r_migrate'),
    Option('lag', list, optional=True, shape=2, datatype=float,
           alternatives='r_lag'),
    Option('dip', list, optional=False, shape=2, datatype=float,
           alternatives='r_dip'),

]

options['sheet'] = element_options + [
    Option('structure', str, optional=False,
           validation_func=lambda x: x in ['massive', 'dip']),
    Option('lens_thickness', float, optional=False),
    Option('dipset_d', float, optional=True),
    Option('dip', list, optional=True, shape=2, datatype=float, default=[0., 0.],
           alternatives='r_dip'),
    Option('azimuth', list, optional=True, shape=2, datatype=float, default=[0., 0.],
           alternatives='r_azimuth'),
]


options['trunc_ellip'] = element_options + erosive_element_options + [
    Option('structure', str, optional=False,
           validation_func=lambda x: x in ['massive', 'dip', 'bulb', 'bulb_l', 'random', 'flat']),
    Option('length', float, optional=False),
    Option('width', float, optional=False),
    Option('depth', float, optional=False),
    Option('el_z', float, optional=False),
    Option('paleoflow', list, optional=False, shape=2, datatype=float,
           alternatives='r_paleoflow'),
    Option('azimuth', list, optional=False, shape=2, datatype=float,
           alternatives='r_azimuth'),
    Option('bulbset_d', float, optional=True),
    Option('te_xyz', list, optional=True, shape=[-1, 3], datatype=float,
           alternatives='ll_te_xyz'),
]


options['ext_par'] = element_options + erosive_element_options + [
    Option('structure', str, optional=False,
           validation_func=lambda x: x in ['massive', 'dip']),
    Option('width', float, optional=False),
    Option('depth', float, optional=False),
    Option('h', float, optional=False),
    Option('k', float, optional=False),
    Option('ds', float, optional=False),
    Option('eps_factor', float, optional=False),
    Option('channel_no', float, optional=False),
]


# Hydraulics options
# ==================
options['hydraulics'] = [
    Option('gen', bool, optional=True, default=True, alternatives='flag_gen'),
    Option('hydro', list, optional=False, shape=[-1], datatype=str,
           alternatives='l_hydro'),
    Option('k_h', list, optional=False, shape='hydro', datatype=float,
           alternatives='r_k_h'),
    Option('sig_y', list, optional=False, shape='hydro', datatype=float,
           alternatives='r_sig_y'),
    Option('k_ratio', list, optional=False, shape='hydro', datatype=float,
           alternatives='r_k_ratio'),
    Option('n', list, optional=False, shape='hydro', datatype=float,
           alternatives='r_n'),
    Option('sig_n', list, optional=False, shape='hydro', datatype=float,
           alternatives='r_sig_n'),
    Option('ycorlengths', list, optional=False, shape=['hydro', 3], datatype=float,
           alternatives='ll_ycorlengths'),
    Option('ncorlengths', list, optional=False, shape=['hydro', 3], datatype=float,
           alternatives='ll_ncorlengths'),
    Option('k_ztrend', list, optional=True, shape=2, datatype=float,
           alternatives='r_k_ztrend'),
    Option('k_xtrend', list, optional=True, shape=2, datatype=float,
           alternatives='r_k_xtrend'),
    Option('n_ztrend', list, optional=True, shape=2, datatype=float,
           alternatives='r_n_ztrend'),
    Option('n_xtrend', list, optional=True, shape=2, datatype=float,
           alternatives='r_n_xtrend'),
]


# Flowtrans options
# =================
options['flowtrans'] = [
    Option('hin', list, optional=True, shape=3, datatype=float),
    Option('hout', list, optional=True, shape=3, datatype=float),
    Option('gradh', float, optional=True),
    Option('q_in', float, optional=True),
    Option('kw_gradh', float, optional=True),
    Option('kw_q_in', float, optional=True)
]



import sys
import hyvr
import argparse

parser = argparse.ArgumentParser(description="HyVR: Turning your geofantasy into reality!")
parser.add_argument('--overwrite', action='store_true',
        help='whether to overwrite the old run directory')
parser.add_argument('filename', default=0, action="store", nargs='?',
        help='HyVR config file. Defaults to a testcase.')
args = parser.parse_args()


if args.overwrite == False:
    args.overwrite = None
try:
    hyvr.run(args.filename, args.overwrite)
except Exception as e:
    print("Something went wrong:", e, file=sys.stderr)
    sys.exit(1)



"""
This file contains functions to convert the internal representation of a HyVR
model output (a dictionary of arrays) to common data or model input formats.

The functions should be named `to_<format>` and take the following parameters:

model : Model instance (see model.py)
    The model object holding all the data arrays
fname : str
    Where to save the file

If you want to add an output format, add a function in the same style and link
the inifile-name of the format with the function and the file extension in the
``create_outputs`` function below (in the dictionary ``output_desc``).
"""

import os
import pickle
import scipy.io as sio
import numpy as np
from hyvr.utils import print_to_stdout, try_makefolder


def create_outputs(model, realization_dir, runname, formats):
    """
    This functions creates output files based on the model output in the
    current run directory.

    This calls mainly the functions defined below for the different datatypes.

    Parameters
    ----------
    model : Model instance (see model.py)
        The model object holding all the data arrays
    realization_dir : str
        Directory where the current realization results should be stored.
    runname : str
        Name of the model run
    formats : list of str
        List of output formats
    """

    # This dictionary links the output format names from the ini-file with output
    # functions and file endings.
    # If you want to add an output format, add a function and add the description here.
    output_desc = {
        'mat':to_mat,
        'py':to_pickle,
        'npz':to_npz,
        'h5':to_hdf5,
        'vtr':to_vtr,
        'mf':to_modflow,
        'mf6':to_mf6,
        'hgs':to_hgs,
    }

    for fmt in formats:
        if fmt not in output_desc:
            raise ValueError('No such output format: ' + fmt)
        fname = os.path.join(os.path.abspath(realization_dir), runname)
        output_desc[fmt](model, fname)
        print_to_stdout('Saved', fmt, 'output to', realization_dir)


def to_mat(model, fname):
    """
    Saves model output as .mat file.

    Parameters
    ----------
    model : Model instance (see model.py)
        The model object holding all the data arrays
    fname : str
        Where to save the file (without file format extension)
    """
    sio.savemat(fname+'.mat', model.data)


def to_pickle(model, fname):
    """
    Saves model output as .pickle file.

    Parameters
    ----------
    model : Model instance (see model.py)
        The model object holding all the data arrays
    fname : str
        Where to save the file (without file format extension)
    """
    with open(fname+'.pickle', 'wb') as outfile:
        pickle.dump(model.data, outfile, protocol=pickle.HIGHEST_PROTOCOL)

def to_npz(model, fname):
    """
    Saves model output as .npz file.

    Parameters
    ----------
    model : Model instance (see model.py)
        The model object holding all the data arrays
    fname : str
        Where to save the file (without file format extension)
    """
    np.savez_compressed(fname+'.npz', **model.data)


def to_hdf5(model, fname):
    """
    Saves model output as .h5 file. This requires h5py.

    Parameters
    ----------
    model : Model instance (see model.py)
        The model object holding all the data arrays
    fname : str
        Where to save the file (without file format extension)
    """
    try:
        import h5py
        with h5py.File(fname+'.h5', 'w') as hf:
            for key in model.data:
                hf.create_dataset(key, data=model.data[key], compression=True)
    except ImportError:
        raise ValueError('h5 output not possible: h5py not found.')


def to_vtr(model, fname):
    """
    Saves model output as .vtr file. This requires pyevtk

    Parameters
    ----------
    model : Model instance (see model.py)
        The model object holding all the data arrays
    fname : str
        Where to save the file (without file format extension)
    """
    # ktensor can not be saved as vtr
    try:
        from pyevtk.hl import gridToVTK
        data_dict = {key:model.data[key] for key in model.data if key != "ktensors"}
        xv = np.arange(model.grid.x0, model.grid.xmax+model.grid.dx, model.grid.dx)
        yv = np.arange(model.grid.y0, model.grid.ymax+model.grid.dy, model.grid.dy)
        zv = np.arange(model.grid.z0, model.grid.zmax+model.grid.dz, model.grid.dz)
        gridToVTK(fname, xv, yv, zv, cellData=data_dict)
    except ImportError:
        raise ValueError('VTR output not possible: pyevtk not found. Remove the vtr output in the *.ini file or install pyevtk.')

def to_modflow(model, fname):
    """
    Saves model output in modflow format. This requires flopy.

    Parameters
    ----------
    model : Model instance (see model.py)
        The model object holding all the data arrays
    fname : str
        Where to save the file (without file format extension)
    """
    try:
        import flopy
    except ImportError:
        raise ValueError('modflow output not possible: flopy not found.')
    # For modflow we want to create a new folder instead of only a file. The folder name is the base
    # name of the passed filename
    realization_dir = os.path.dirname(fname)
    runname = os.path.basename(fname)
    mfdir = os.path.join(realization_dir, 'MODFLOW')
    mfname = os.path.join(mfdir, runname)
    try_makefolder(mfdir)

    # Assign name and create modflow model object
    mf = flopy.modflow.Modflow(mfname, exe_name='mf2005')

    # Create the discretization object
    ztop = model.grid.z0 + model.grid.lz
    zbot = model.grid.z0
    botm = np.linspace(ztop, zbot, model.grid.nz + 1)
    dis = flopy.modflow.ModflowDis(mf, model.grid.nz, model.grid.nx, model.grid.ny, 
                                   delr=model.grid.dx, delc=model.grid.dy,
                                   top=ztop, botm=botm[1:])

    # Variables for the BAS package
    ibound = np.ones((model.grid.nz, model.grid.nx, model.grid.ny), dtype=np.int32)
    ibound[:, :, 0] = -1
    ibound[:, :, -1] = -1

    strt = np.ones((model.grid.nz, model.grid.nx, model.grid.ny), dtype=np.float32)
    strt[:, :, 0] = model.flowtrans['hin'][0]
    strt[:, :, -1] = model.flowtrans['hout'][0]

    bas = flopy.modflow.ModflowBas(mf, ibound=ibound, strt=strt)

    # Assign hydraulic conductivity
    hyvr_hk = np.transpose(model.data['k_iso'], (2, 0, 1))
    hyvr_layvka = 1                                           # VKA dataset is ratio of horizontal K
    if 'anirat' in model.data.keys():
        hyvr_vka = np.transpose(model.data['anirat'], (2, 0, 1))

        # Add LPF package to the MODFLOW model
        lpf = flopy.modflow.ModflowLpf(mf,                        # Modflow object
                                       hk=hyvr_hk,              # Horizontal hydraulic conductivity
                                       layvka=hyvr_layvka,      # Flag for each layer of anisotropic ratio
                                       vka=hyvr_vka)            # Anisotropy ratios.
    else:
        # Add LPF package to the MODFLOW model
        lpf = flopy.modflow.ModflowLpf(mf,                        # Modflow object
                                       hk=hyvr_hk)              # Horizontal hydraulic conductivity

    oc = flopy.modflow.ModflowOc(mf)        # Add OC package to the MODFLOW model
    pcg = flopy.modflow.ModflowPcg(mf)      # Add PCG package to the MODFLOW model
    mf.write_input()                        # Write the MODFLOW model input files
    



def to_mf6(model, fname):
    """
    Saves model output in mf6 format. This requires flopy.

    Parameters
    ----------
    model : Model instance (see model.py)
        The model object holding all the data arrays
    fname : str
        Where to save the file (without file format extension)
    """

    try:
        import flopy
    except ImportError:
        raise ValueError('mf6 output not possible: Flopy not installed.')

    # For modflow we want to create a new folder instead of only a file. The folder name is the base
    # name of the passed filename
    realization_dir = os.path.dirname(fname)
    runname = os.path.basename(fname)
    mfdir = os.path.join(realization_dir, 'mf6')
    mfname = os.path.join(mfdir, runname)
    try_makefolder(mfdir)

    # Transpose HyVR arrays for MF6 input
    transpose_order = (2, 1, 0)
    k_iso = np.transpose(model.data['k_iso'], transpose_order)
    if set(['anirat', 'dip', 'azim']).issubset(model.data.keys()):
        anirat = np.transpose(model.data['anirat'], transpose_order)
        dip = np.transpose(model.data['dip'], transpose_order)
        azim = np.transpose(model.data['azim'], transpose_order)
        xt3d = True
    else:
        xt3d = False

    """ create simulation """
    sim = flopy.mf6.MFSimulation(sim_name=runname,
                                 version='mf6',
                                 exe_name='mf6',
                                 sim_ws=mfdir)
                                 # sim_tdis_file='simulation.tdis')

    """ Create the Flopy temporal discretization object - STEADY-STATE """
    tdis = flopy.mf6.modflow.mftdis.ModflowTdis(sim,
                                                time_units='DAYS')

    """ create gwf model """
    gwf = flopy.mf6.MFModel(sim, modelname=runname)


    ims = flopy.mf6.ModflowIms(sim,
                               print_option='SUMMARY',
                               complexity='COMPLEX',
                               outer_hclose=1e-3,
                               outer_maximum=500,
                               under_relaxation='NONE',
                               inner_maximum=100,
                               inner_hclose=1e-4,
                               rcloserecord=0.001,
                               linear_acceleration='BICGSTAB',
                               scaling_method='NONE',
                               reordering_method='NONE',
                               relaxation_factor=0.97)
    sim.register_ims_package(ims, [gwf.name])

    """ Create discretization """
    ztop = model.grid.z0 + model.grid.lz
    zbot = model.grid.z0
    botm = np.around(np.arange(ztop, zbot-model.grid.dz, -model.grid.dz), decimals=3)
    dis = flopy.mf6.modflow.mfgwfdis.ModflowGwfdis(gwf,
                                                   nlay=model.grid.nz,
                                                   nrow=model.grid.ny,
                                                   ncol=model.grid.nx,
                                                   delr=model.grid.dy,
                                                   delc=model.grid.dx,
                                                   top=ztop,
                                                   botm=botm[1:],
                                                   fname='{}.dis'.format(runname))

    """ Create Node Property Flow package object """
    if xt3d is True:
        npf_package = flopy.mf6.ModflowGwfnpf(gwf,
                                              save_flows=True, icelltype=0, xt3doptions='',
                                              k=k_iso,                                  # within-bedding hydraulic conductivity
                                              k33=k_iso/anirat,                         # across-bedding hydraulic conductivity
                                              angle1=azim,                              # azimuth
                                              angle2=dip,                               # dip
                                              angle3=np.zeros((model.grid.nz, model.grid.ny, model.grid.nx)))   # no rotation
    else:
        npf_package = flopy.mf6.ModflowGwfnpf(gwf,
                                              save_flows=True, icelltype=0,
                                              k=k_iso)                                # within-bedding hydraulic conductivity)

    """ Create constant head package """
    if model.flowtrans['hin'] is not None:
        hin = model.flowtrans['hin'][0]
        hout = model.flowtrans['hout'][0]

    elif flowtrans['gradh'] is not None:
        hout = 1
        hin = hout + model.grid.lx * flowtrans['gradh']

    if np.any([model.flowtrans['hin'] is not None, model.flowtrans['gradh'] is not None]):
        chd_rec = []
        for layer in range(0, model.grid.nz):
            for row in range(0, model.grid.ny):
                chd_rec.append(((layer, row, 0), hin))         # Apply at model inlet
                chd_rec.append(((layer, row, model.grid.nx-1), hout))   # Apply at model outlet
        # chd = flopy.mf6.modflow.mfgwfchd.ModflowGwfchd(gwf, maxbound=len(chd_rec),
        #                                                stress_period_data=chd_rec, save_flows=True)

        chd = flopy.mf6.modflow.mfgwfchd.ModflowGwfchd(gwf, maxbound=len(chd_rec),
                                                       stress_period_data=chd_rec, save_flows=True)
    elif model.flowtrans['q_in'] is not None:
        """ Apply fixed head at model outlet if fixed head at inlet"""
        hin = 1
        hout = 1
        chd_rec = []
        for layer in range(0, model.grid.nz):
            for row in range(0, model.grid.ny):
                chd_rec.append(((layer, row, model.grid.nx-1), hout))   # Apply at model outlet

        chd = flopy.mf6.modflow.mfgwfchd.ModflowGwfchd(gwf, maxbound=len(chd_rec),
                                                       stress_period_data=chd_rec, save_flows=True)
    else:
        hin = 1
        hout = 1

    """ Create the initial conditions package """
    # Create linear initial condition
    # hstart = np.ones_like(k_iso) *(hin - hout)/2
    hstart = np.ones_like(k_iso) * np.linspace(hin, hout, model.grid.nx)
    ic = flopy.mf6.modflow.mfgwfic.ModflowGwfic(gwf, strt=hstart)



    """ Create well package """
    # Apply constant discharges at model faces
    if 'q_in' in model.flowtrans:
        if model.flowtrans['q_in'] is not None:
            q_in = model.flowtrans['q_in']
        else:
            q_in = 0.01

        # if 'q_out' in flowtrans:
        #     q_out = flowtrans['q_out']
        # else:
        #     q_out = -q_in

        wel_rec = []
        for layer in range(0, model.grid.nz):
            for row in range(0, model.grid.ny):
                wel_rec.append(((layer, row, 0),  q_in, 'inlet'))          # Apply at model inlet
                # wel_rec.append(((layer, row, model.nx-1), q_out, 'outlet'))        # Apply at model outlet

        # Apply to model
        wel = flopy.mf6.ModflowGwfwel(gwf,
                                      print_input=True,
                                      print_flows=True,
                                      save_flows=True,
                                      boundnames=True,
                                      maxbound=len(wel_rec),
                                      stress_period_data=wel_rec)


    """ Create the output control package """
    headfile = '{}.hds'.format(runname)
    head_filerecord = [headfile]
    budgetfile = '{}.cbc'.format(runname)
    budget_filerecord = [budgetfile]
    saverecord = [('HEAD', 'ALL'),
                  ('BUDGET', 'ALL')]
    printrecord = [('HEAD', 'LAST')]
    oc = flopy.mf6.modflow.mfgwfoc.ModflowGwfoc(gwf,
                                                saverecord=saverecord,
                                                head_filerecord=head_filerecord,
                                                budget_filerecord=budget_filerecord,
                                                printrecord=printrecord)

    # write simulation
    sim.write_simulation()


def to_hgs(model, fname):
    """
    Saves model output in the HydroGeoSphere format.

    Parameters
    ----------
    model : Model instance (see model.py)
        The model object holding all the data arrays
    fname : str
        Where to save the file (without file format extension)
    """
    realization_dir = os.path.dirname(fname)
    runname = os.path.basename(fname)
    hgsdir = os.path.join(realization_dir, 'HGS')
    try_makefolder(hgsdir)

    uid = np.arange(1, len(model.data['ktensors'][:, :, :, 1, 2].flatten()) + 1)    # Create list of IDs
    vals_to_write = {'ktensors': np.column_stack((uid,
                                      model.data['ktensors'][:, :, :, 0, 0].flatten(),            # K_xx
                                      model.data['ktensors'][:, :, :, 1, 1].flatten(),            # K_yy
                                      model.data['ktensors'][:, :, :, 2, 2].flatten(),            # K_zz
                                      model.data['ktensors'][:, :, :, 0, 1].flatten(),            # K_xy
                                      model.data['ktensors'][:, :, :, 0, 2].flatten(),            # K_xz
                                      model.data['ktensors'][:, :, :, 1, 2].flatten())),          # K_yz
                     'porosity': np.column_stack((uid,
                                      model.data['poros'].flatten()))}
    val_fmts = {'ktensors': '%u %1.3e %1.3e %1.3e %1.3e %1.3e %1.3e',
                'porosity': '%u %1.3f'}

    # Loop over properties to write
    for val in vals_to_write:
        val_filepath = os.path.join(hgsdir, val + '.txt')       # File name of HGS output file
        np.savetxt(val_filepath, vals_to_write[val], fmt=val_fmts[val])



import sys
import numpy as np
import matplotlib.pyplot as plt

def cross_section_pcolor(model, field, y=None, x=None, log=False, xlim_y=None, ylim_y=None,
        xlim_x=None, ylim_x=None, cmap=None):
    """
    Create a pcolor plot of a cross section of the specified field.

    Parameters
    ----------
    model : Model object
    field : str
        The field to be shown, e.g. 'ae', 'fac', 'hat', 'k_iso', etc.
    y : sequence of values or None, optional (default: None)
        If a sequence of values or a single value is given, cross section at this y-value will be
        plotted.
        If it is None, a y-cross section through the center of the model domain will be plotted.
        If it is an empty list, no y-cross section will be plotted.
    x : sequence of values or None, optional (default: None)
        If a sequence of values or a single value is given, cross section at this x-value will be
        plotted.
        If it is None, a x-cross section through the center of the model domain will be plotted.
        If it is an empty list, no x-cross section will be plotted.
    log : bool, optional (default: False)
        Whether to plot the logarithm of the field.
    xlim_y, ylim_y : lists/tuples, optional (default: None)
        Axis limits for the y-cross-sections
    xlim_x, ylim_x : lists/tuples, optional (default: None)
        Axis limits for the x-cross-sections
    cmap : matplotlib colormap, optional (default: None)
        Colormap to use for the plot, e.g. 'prism'.
    """
    if y is None:
        y = [model.grid.y0 + model.grid.ly/2]
    if x is None:
        x = [model.grid.x0 + model.grid.lx/2]
    if np.isscalar(y):
        y = [y]
    if np.isscalar(x):
        x = [x]

    for yi in y:
        fig, ax = plt.subplots()
        y_index = int(np.round((yi-(model.grid.y0+model.grid.dy/2)/model.grid.dy)))
        if 0 <= y_index and y_index < model.grid.ny:
            im = ax.pcolor(model.grid.X[:,y_index,:].T, model.grid.Z[:,y_index,:].T,
                    model.data[field][:,y_index,:].T, cmap=cmap)
            ax.set_title(field + ', y = {:.2f}'.format(yi))
            ax.set_xlabel('x')
            ax.set_ylabel('z')
            if xlim_y is not None: ax.set_xlim(xlim)
            if ylim_y is not None: ax.set_ylim(ylim)
            ax.set_aspect('equal')
            fig.colorbar(im, ax=ax)
            plt.show(fig)
        else:
            print("Warning: y = {:.2f} not in model domain".format(yi), file=sys.stderr)

    for xi in x:
        fig, ax = plt.subplots()
        x_index = int(np.round((xi-(model.grid.x0+model.grid.dx/2)/model.grid.dx)))
        if 0 <= x_index and x_index < model.grid.nx:
            im = ax.pcolor(model.grid.Y[x_index,:,:].T, model.grid.Z[x_index,:,:].T,
                    model.data[field][x_index,:,:].T, cmap=cmap)
            ax.set_title(field + ', x = {:.2f}'.format(xi))
            ax.set_xlabel('y')
            ax.set_ylabel('z')
            if xlim_x is not None: ax.set_xlim(xlim)
            if ylim_x is not None: ax.set_ylim(ylim)
            ax.set_aspect('equal')
            fig.colorbar(im, ax=ax)
            plt.show(fig)
        else:
            print("Warning: x = {:.2f} not in model domain".format(xi), file=sys.stderr)







"""
This module contains optimized versions of routines that were once in sim or utils.

:Author: Samuel Scherrer
"""

import numpy as np
import math
cimport numpy as np
cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
def reindex(inray):
    """
    Reindex array from 0

    Parameters:
        inray:		Array with indices

    Returns:
        inray:      Array with altered indices. ATTENTION: This changes ``inray``!!

    """

    uniq = np.unique(inray)
    remat = dict(zip(uniq, range(len(uniq))))
    linear_inray = inray.reshape(inray.size)
    for i in range(len(linear_inray)):
        linear_inray[i] = remat[linear_inray[i]] + 1
    return inray


@cython.boundscheck(False)
@cython.wraparound(False)
def select_trough(np.ndarray[np.float_t, ndim=3] Xd,
                  np.ndarray[np.float_t, ndim=3] Yd,
                  np.ndarray[np.float_t, ndim=3] Zd,
                  np.float_t a=1.0, np.float_t b=1.0, np.float_t c=1.0,
                  np.float_t alpha=0.0):
    """
    Select cells that belong to the trough.

    Parameters
    ----------
    Xd, Yd, Zd : np.ndarray[np.float_t, ndim=3]
        x, y, and z distances of grid cells to trough center
    a, b, c : np.float
        Trough parameters (principal semi-axes lengths)
    alpha : np.float
        Rotation angle in degree

    Returns
    -------
    select : np.ndarray[np.bool, ndim=3]
        Mask that selects only points inside the trough
    """
    cdef np.float_t xi, yi, zi, cos, sin, alpha_rad, r2, max_ab
    cdef np.int_t nx, ny, nz, i, j, k
    nx = Xd.shape[0]
    ny = Xd.shape[1]
    nz = Xd.shape[2]

    # calculate cos and sin
    alpha_rad = alpha/180*np.pi
    cos = np.cos(alpha_rad)
    sin = np.sin(alpha_rad)

    # The trough is at most max(a, b) from the center
    max_ab = max(a, b)

    # assign mask
    cdef np.ndarray[np.uint8_t, ndim=3, cast=True] select = np.empty((nx, ny, nz), dtype=np.bool)
    for i in range(nx):
        for j in range(ny):
            for k in range(nz):
                xi = Xd[i,j,k]
                yi = Yd[i,j,k]
                zi = Zd[i,j,k]
                if zi <= 0 and abs(xi) <= max_ab and abs(yi) <= max_ab:
                    r2 = (xi**2*cos**2 + 2*xi*yi*cos*sin + yi**2*sin**2)/a**2
                    r2 += (xi**2*sin**2 - 2*xi*yi*cos*sin + yi**2*cos**2)/b**2
                    r2 += zi**2/c**2
                    select[i,j,k] = (r2 <= 1)
                else:
                    select[i,j,k] = 0

    return select



@cython.boundscheck(False)
@cython.wraparound(False)
def set_anisotropic_ktensor(np.ndarray[np.float_t, ndim=5] ktensors,
            np.ndarray[np.float_t, ndim=3] k_iso,
            np.ndarray[np.float_t, ndim=3] azim,
            np.ndarray[np.float_t, ndim=3] dip,
            np.ndarray[np.float_t, ndim=3] anirat):
    """
    Calculate and set the anisotropic conductivity tensor for every grid point

    Parameters
    ----------
    ktensors : np.ndarray, dtype = np.float, shape = (nx, ny, nz, 3, 3)
        The matrix of conductivity tensors. This matrix will be altered in place.
    k_iso : np.ndarray, dtype = np.float, shape = (nx, ny, nz)
        The matrix of isotropic hydraulic conductivities
    azim : np.ndarray, dtype = np.float, shape = (nx, ny, nz)
        The matrix of azimuth angles (in radians)
    dip : np.ndarray, dtype = np.float, shape = (nx, ny, nz)
        The matrix of dip angles (in radians)
    anirat : np.ndarray, dtype = np.float, shape = (nx, ny, nz)
        The matrix of anisotropy ratios

    Returns
    -------
    None
    """
    cdef np.float_t kappa, psi, a, sink, cosk, sinp, cosp
    cdef np.float_t a11, a12, a13, a22, a23, a33
    cdef np.float_t kiso
    cdef np.int_t imax, jmax, kmax, i, j, k
    imax = k_iso.shape[0]
    jmax = k_iso.shape[1]
    kmax = k_iso.shape[2]
    for i in range(imax):
        for j in range(jmax):
            for k in range(kmax):
                kappa = azim[i,j,k]
                psi = dip[i,j,k]
                a = 1/anirat[i,j,k]
                sink = np.sin(kappa)
                cosk = np.cos(kappa)
                sinp = np.sin(psi)
                cosp = np.cos(psi)
                kiso = k_iso[i,j,k]
                a00 = kiso*(sink**2 + cosk**2*(cosp**2 + a*sinp**2))
                a01 = kiso*(1-a)*sink*cosk*sinp**2
                a02 = kiso*((a-1)*(np.sin(kappa+2*psi)-np.sin(kappa-2*psi)))/4
                a11 = kiso*((a-1)*sink**2*sinp**2+1)
                a12 = kiso*((1-a)*(np.cos(kappa-2*psi)-np.cos(kappa+2*psi)))/4
                a22 = kiso*(a*cosp**2+sinp**2)

                ktensors[i,j,k,0,0] = a00
                ktensors[i,j,k,0,1] = a01
                ktensors[i,j,k,0,2] = a02
                ktensors[i,j,k,1,0] = a01
                ktensors[i,j,k,1,1] = a11
                ktensors[i,j,k,1,2] = a12
                ktensors[i,j,k,2,0] = a02
                ktensors[i,j,k,2,1] = a12
                ktensors[i,j,k,2,2] = a22

# @cython.boundscheck(False)
# @cython.wraparound(False)
# def assign_between(np.ndarray[np.int16_t, ndim=3] strata_number,
                   # np.ndarray[np.float_t, ndim=2] cs_z_bottom,
                   # np.ndarray[np.float_t, ndim=2] cs_z_top,
                   # np.int_t n_strata,
                   # np.float_t z0, np.float_t dz):
    # """
    # Assigns all points below the contact surface to the strata by decreasing
    # its strata number by 1.

    # Parameters
    # ----------
    # strata_number : np.npdarray[np.int16_t, ndim=3]
        # Array of strata numbers. This will be altered.
    # cs_z_bottom : np.ndarray[np.float_t, ndim=2]
        # z-coordinate of the bottom contact surface at every point in the x-y-plane
    # cs_z : np.ndarray[np.float_t, ndim=2]
        # z-coordinate of the top contact surface at every point in the x-y-plane
    # n_strata : int
        # number that should be assigned
    # z0: np.float
        # z-coordinate of the lowest grid layer
    # dz : np.float
        # distance between grid layers
    # """
    # cdef np.int_t imax, jmax, kmax, below_idx, above_idx, i, j
    # cdef np.int16_t n_strata_16
    # imax = strata_number.shape[0]
    # jmax = strata_number.shape[1]
    # kmax = strata_number.shape[2]
    # n_strata_16 = np.int16(n_strata)
    # for i in range(imax):
        # for j in range(jmax):
            # z_bottom = cs_z_bottom[i,j]
            # z_top = cs_z_top[i,j]
            # # find index that is below/above contact surface
            # bottom_idx = int(math.ceil((z_bottom - (z0+0.5*dz))/dz))
            # top_idx = int(math.floor((z_top - (z0+0.5*dz))/dz))
            # for k in range(bottom_idx,top_idx+1):
                # strata_number[i,j,k] = n_strata_16



# @cython.boundscheck(False)
# @cython.wraparound(False)
# def trough_normalized_distance_squared(np.float_t x, np.float_t y, np.float_t z,
#                                        np.float_t a, np.float_t b, np.float_t c,
#                                        np.float_t sin, np.float_t cos):
#     return ((x*cos + y*sin)/a)**2 + ((-x*sin+y*cos)/b)**2 + (z/c)**2

# @cython.boundscheck(False)
# @cython.wraparound(False)
# def trough_normalized_vector(np.float_t x, np.float_t y, np.float_t z,
#                              np.float_t a, np.float_t b, np.float_t c,
#                              np.float_t sin, np.float_t cos):
#     cdef np.ndarray[np.float_t, ndim=1] d = np.empty(3, dtype=np.float)
#     d[0] = (x*cos + y*sin)/a
#     d[1] = (-x*sin + y*cos)/b
#     d[2] = z/c
#     return d


cdef int sign(double x):
    return (x > 0) - (x < 0)



@cython.boundscheck(False)
@cython.wraparound(False)
def curve_interp(double [:] xc, double [:] yc, double spacing):
    """
    Interpolate evenly spaced points along a curve. This code is based on code in an answer posted by 'Unutbu' on
    http://stackoverflow.com/questions/19117660/how-to-generate-equispaced-interpolating-values (retrieved 17/04/2017)

    Parameters:
        xc:			x coordinates of curve
        yc:			y coordinates of curve
        spacing:	Spacing between points

    Returns:
        - xn - x coordinates of interpolated points
        - yn - y coordinates of interpolated points

    """
    cdef int ic, j, i
    cdef double total_dist, tol
    cdef double [:] t
    cdef double [:] xn
    cdef double [:] yn
    cdef list idx

    t = np.arange(xc[0], len(xc), spacing * 0.1)
    xc = np.interp(t, np.arange(len(xc)), xc)
    yc = np.interp(t, np.arange(len(yc)), yc)
    tol = spacing
    ic, idx = 0, [0]
    while ic < len(xc):
        total_dist = 0
        j = ic + 1 # this is to make sure j is initialized
        for j in range(ic+1, len(xc)):
            total_dist += math.sqrt((xc[j] - xc[j-1]) ** 2 + (yc[j] - yc[j-1]) ** 2)
            if total_dist > tol:
                idx.append(j)
                break
        ic = j + 1

    xn = np.zeros(len(idx))
    yn = np.zeros(len(idx))
    for i, ic in enumerate(idx):
        xn[i] = xc[ic]
        yn[i] = yc[ic]

    return xn, yn



"""
This module contains all options for the ini-files. If you want to add an option, add it here and potentially do some post-processing in parameters.

:Authors: Samuel Scherrer, Jeremy P. Bennett
"""


from hyvr.input.option_parsing import Option

options = {}

# Run options
# ===========
options['run'] = [
    Option('runname', str, optional=True, default=None),
    Option('modeldir', str, optional=True, default=None),
    Option('numsim', int, optional=True, default=1),
    Option('outputs', list, optional=True, shape=[-1], default=[], datatype=str,
           validation_func=lambda x: x in ['vtr', 'py', 'mat', 'npz', 'h5', 'mf', 'mf6', 'hgs']),
    Option('overwrite_old_output', bool, optional=True, default=False),
]


# Model options
# =============
options['model'] = [
    Option('x0', float, optional=True, default=0.0),
    Option('y0', float, optional=True, default=0.0),
    Option('z0', float, optional=True, default=0.0),
    Option('dx', float, optional=False),
    Option('dy', float, optional=True), # use value of 'dx' if 'dy' is missing
    Option('dz', float, optional=True),
    Option('lx', float, optional=False),
    Option('ly', float, optional=False),
    Option('lz', float, optional=False),
    Option('periodic', bool, optional=True, default=False),
    Option('hydraulics', bool, optional=True, default=True),
    Option('anisotropy', bool, optional=True, default=True),
    Option('heterogeneity', bool, optional=True, default=True),
    Option('heterogeneity_level', str, optional=True, default='internal',
           validation_func=lambda x: x in ['facies', 'internal']),
]


# Strata options
# ==============
options['strata'] = [
    Option('strata', list, optional=False, shape=[-1], datatype=str),
    Option('strata_contact_models', list, optional=False, shape=[-1, -1], datatype=str),
    Option('ae', list, optional=False, shape=[-1], datatype=str),
    Option('ae_in_strata', list, optional=False, shape=['strata', -1], datatype=str),
    Option('ae_prob', list, optional=False, shape='ae_in_strata', datatype=float),
    Option('ae_z_mean', list, optional=False, shape='ae_in_strata', datatype=float),
    Option('avul_prob', list, optional=False, shape=['strata', 1], datatype=float),
    Option('avul', list, optional=False, shape=['strata', 2], datatype=float),
    Option('bg_facies', list, optional=False, shape='strata', datatype=str),
    Option('bg_azim', list, optional=False, shape='strata', datatype=float),
    Option('bg_dip', list, optional=False, shape='strata', datatype=float),
]


# Element options
# ===============
element_options = [
    Option('geometry', str, optional=False,
           validation_func=lambda x: x in ['trough', 'channel', 'sheet']),
    Option('contact_model', list, optional=True, shape=[-1], datatype=str),
    Option('facies', list, optional=False, shape=[-1], datatype=str),
    Option('altfacies', list, optional=True, shape=['facies', -1], datatype=str),
    Option('size_ztrend', list, optional=True, shape=2, datatype=float),
    Option('k_ztrend', list, optional=True, shape=2, datatype=float),
    Option('k_xtrend', list, optional=True, shape=2, datatype=float),
    Option('n_xtrend', list, optional=True, shape=2, datatype=float),
    Option('n_ztrend', list, optional=True, shape=2, datatype=float),
    Option('dip', list, optional=True, shape=2, default=[0.0, 0.0], datatype=float),
    Option('azimuth', list, optional=True, shape=2, datatype=float, default=[0., 0.]),
    # These are only necessary for the erosive elements, but they are not used for sheets anyways.
    Option('bg_facies', str, optional=True, default=None),
    Option('bg_azim', float, optional=True, default=float('NaN')),
    Option('bg_dip', float, optional=True, default=float('NaN')),
]

erosive_element_options = [
    Option('agg', float, optional=False),
    Option('buffer', float, optional=True, default=0.0),
    Option('lag', list, optional=True, shape=2, datatype=float),
    Option('lag_height', float, optional=True, default=0.0),
    Option('lag_facies', str, optional=True, default=None),
]

options['sheet'] = element_options + [
    Option('structure', str, optional=False,
           validation_func=lambda x: x in ['massive', 'dip']),
    Option('dipset_dist', float, optional=True),
    Option('lens_thickness', float, optional=False),
]


options['trough'] = element_options + erosive_element_options + [
    Option('structure', str, optional=False,
           validation_func=lambda x: x in ['massive', 'dip', 'bulb', 'bulb_sets', 'random', 'flat']),
    Option('dipset_dist', float, optional=True),
    Option('bulbset_dist', float, optional=True),
    Option('length', float, optional=False),
    Option('width', float, optional=False),
    Option('depth', float, optional=False),
    Option('trough_density', float, optional=False),
    Option('paleoflow', list, optional=False, shape=2, datatype=float),
    Option('te_xyz', list, optional=True, shape=[-1, 3], datatype=float),
    Option('migrate', list, optional=True, shape=4, datatype=float),
]


options['channel'] = element_options + erosive_element_options + [
    Option('structure', str, optional=False,
           validation_func=lambda x: x in ['massive', 'dip']),
    Option('dipset_dist', float, optional=True),
    Option('width', float, optional=False),
    Option('depth', float, optional=False),
    Option('h', float, optional=False),
    Option('k', float, optional=False),
    Option('ds', float, optional=False),
    Option('eps_factor', float, optional=False),
    Option('channel_no', int, optional=False),
    Option('mig', float, optional=True, datatype=float, default=0.),
    Option('flow_angle', float, optional=True, default=0.),
]


# Hydraulics options
# ==================
options['hydraulics'] = [
    Option('hydrofacies', list, optional=False, shape=[-1], datatype=str),
    Option('k_h', list, optional=False, shape='hydrofacies', datatype=float),
    Option('sig_y', list, optional=False, shape='hydrofacies', datatype=float),
    Option('k_ratio', list, optional=False, shape='hydrofacies', datatype=float),
    Option('n', list, optional=False, shape='hydrofacies', datatype=float),
    Option('sig_n', list, optional=False, shape='hydrofacies', datatype=float),
    Option('ycorlengths', list, optional=False, shape=['hydrofacies', 3], datatype=float),
    Option('ncorlengths', list, optional=False, shape=['hydrofacies', 3], datatype=float),
    Option('k_ztrend', list, optional=True, shape=2, datatype=float),
    Option('k_xtrend', list, optional=True, shape=2, datatype=float),
    Option('n_ztrend', list, optional=True, shape=2, datatype=float),
    Option('n_xtrend', list, optional=True, shape=2, datatype=float),
]


# Flowtrans options
# =================
options['flowtrans'] = [
    Option('hin', list, optional=True, shape=3, datatype=float),
    Option('hout', list, optional=True, shape=3, datatype=float),
    Option('gradh', float, optional=True),
    Option('q_in', float, optional=True),
    Option('kw_gradh', float, optional=True),
    Option('kw_q_in', float, optional=True)
]



"""
This module contains the architectural element types.
"""

from hyvr.classes.ae_realizations import *

class AEType:
    """
    Architectural element types are the kind of things that are read from the
    supplied ini-file. For example, in the ``made.ini`` testcase, there are the
    architectural element types "clay_sheet", "sand_sheet", "clay_lens",
    "crossbedded_scour", ...
    In HyVR, these are instances of the ``AEType`` class.

    The actual AEs that are realized in a HyVR simulation are instances of
    ``AERealization`` and are implemented in 'ae_realization.py'.

    To generate a realization, you should normally call the
    ``generate_realization`` method of a type instance.
    """

    def __init__(self, ae_section, name):
        """
        AE type constructor.

        Parameters
        ----------
        ae_section : dict
            Parsed element section from the ini-file.
        name : str
            The name of the element type (e.g. 'crossbedded_scour').
        """
        self.params = ae_section
        self.geometry = ae_section['geometry']
        self.name = name

    def generate_realization(self, bottom_surface, top_surface, stratum, grid):
        """
        Generates a realization of the AE type.

        For example, if the AE geometry is 'trunc_ellip', this generates a
        ``TruncEllip`` object, i.e. a sub-layer containing trough features
        within a stratum.

        Parameters
        ----------
        bottom_surface : ContactSurface object
            The bottom surface of the AE.
        top_surface : ContactSurface object
            The top surface of the AE.
        model : Model object
        """
        if self.geometry == 'trough':
            return TruncEllipAE(bottom_surface, top_surface, self.name, self.params, stratum, grid)
        elif self.geometry == 'channel':
            return ExtParAE(bottom_surface, top_surface, self.name, self.params, stratum, grid)
        elif self.geometry == 'sheet':
            return SheetAE(bottom_surface, top_surface, self.name, self.params, stratum, grid)
        else:
            raise ValueError('No such geometry in ' + self.name + ': ' + self.geometry)









"""
This class holds the abstraction of a HyVR model.
It can be broadly structured in two parts:

* **Model description**: This consists of all info about the model that is read
  from the ini-file.
* **Model realization**: This consists of several 3-d arrays which hold the
    data for the different variables on the grid. This is what get's put out in
    the end. The data is stored in a dictionary of arrays.
"""

import numpy as np
import hyvr.utils as hu
import hyvr.optimized as ho
from hyvr.classes.contact_surface import ContactSurface
from hyvr.classes.contact_surface_utils import *
from hyvr.classes.stratum import Stratum
from hyvr.classes.ae_types import AEType
from hyvr.classes.grid import Grid


class Model:

    def __init__(self, model_dict, strata_dict, elements, flowtrans_section):
        """
        Creates the model object based on the parsed parameter file.

        Parameters
        ----------
        model_dict : dictionary
            The parsed 'model' section of the parameter-file.
        strata_dict : dictionary
            The parsed 'strata' section of the parameter-file.
        elements : dictionary of dictionaries
            Dictionary containing the different element sections.
        flowtrans_section : dictionary
            The parsed 'flowtrans' section of the parameter-file.
        self.data = {}
        """
        self.grid = Grid(
                model_dict['x0'],
                model_dict['y0'],
                model_dict['z0'],
                model_dict['dx'],
                model_dict['dy'],
                model_dict['dz'],
                model_dict['lx'],
                model_dict['ly'],
                model_dict['lz'],
                model_dict['periodic'],
                )
        self.generate_ae_types(elements)

        self.strata_dict = strata_dict

        # anisotropy and heterogeneity
        self.generate_anisotropy = model_dict['anisotropy']
        self.generate_hydraulics = model_dict['hydraulics']
        self.generate_heterogeneity = model_dict['heterogeneity']
        self.heterogeneity_level = model_dict['heterogeneity_level']

        # flowtrans is necessary for model output
        self.flowtrans = flowtrans_section

        self.data = {}



    # STRATA AND ARCHITECTURAL ELEMENT GENERATION
    # ===================================================================================

    def generate_ae_types(self, elements):
        """
        Read in elements sections and generate type objects
        """
        # TODO: This could maybe be moved to model setup
        self.ae_types = {}
        for elem in elements:
            self.ae_types[elem] = AEType(elements[elem], elem)


    def generate_model(self):
        """
        Generates the model.

        This function generates the full model by creating all strata, the AEs
        within the strata and the geometrical objects within these.

        The method starts from the domain top and calls the strata generation
        function for each stratum, which handles the generation of
        architectural elements.
        """
        hu.print_to_stdout('Generating model...')
        strata_dict = self.strata_dict
        self.strata = []
        self.n_strata = len(strata_dict['strata'])

        # These are the additional parameters that a stratum needs
        param_names = [
            'bg_facies',
            'bg_azim',
            'bg_dip',
            'ae_prob',
            'ae_z_mean',
            'avul',
            'avul_prob',
            'strata',
        ]

        top_surface = ContactSurface(self.grid, mode='flat', z=self.grid.lz)
        for si in range(self.n_strata):
            # create top surface
            if si == self.n_strata - 1:
                bottom_surface = ContactSurface(self.grid, mode='flat', z=self.grid.z0)
            else:
                bottom_surface = ContactSurface(self.grid, **strata_dict['contact_models'][-1-si])
            # if the bottom surface is above the top surface, we will assign
            # the top surface values instead
            use_lower_surface_value(bottom_surface, top_surface)

            stratum_params = {name:strata_dict[name][-1-si] for name in param_names}
            ae_types = [self.ae_types[ae_type] for ae_type in strata_dict['ae_in_strata'][-1-si]]


            stratum = Stratum(bottom_surface,
                              top_surface,
                              stratum_params,
                              ae_types,
                              self.grid,
            )

            self.strata.append(stratum)

            # use old top as bottom
            top_surface = bottom_surface


        # enumerate all objects and AEs
        num_ae = 0
        num_ha = 0
        for stratum in self.strata:
            for ae in stratum.aes:
                ae.num = num_ae
                num_ae += 1
                for obj in ae.objects:
                    obj.num_ha = num_ha
                    num_ha += 1




    def generate_hydraulic_parameters(self, hydraulics):
        """
        Generates (heterogeneous) hydraulic parameters

        Parameters
        ----------
        hydraulics : dictionary
            parsed hydraulics section of the ini-file
        """

        hu.print_to_stdout('Generating hydraulic parameters')

        het = self.generate_heterogeneity
        ani = self.generate_anisotropy

        ha_arr = self.data['ha']
        fac = self.data['facies']
        ae_arr = self.data['ae']
        azim = self.data['azim']
        dip = self.data['dip']

        # Initialise storage arrays:
        # hydraulic conductivity
        self.data['k_iso'] = np.zeros((self.grid.nx, self.grid.ny, self.grid.nz),
                                      dtype=np.float)
        # porosity
        self.data['poros' ]= np.zeros((self.grid.nx, self.grid.ny, self.grid.nz),
                                      dtype=np.float)
        # anisotropy ratio
        self.data['anirat' ]= np.ones((self.grid.nx, self.grid.ny, self.grid.nz),
                                      dtype=np.float)

        k_iso = self.data['k_iso']
        poros = self.data['poros']
        anirat = self.data['anirat']

        if het:
            # Heterogeneous case
            for mi in np.unique(ha_arr):
                for fi in np.unique(fac[ha_arr == mi]):
                    mifi = (ha_arr == mi) & (fac == fi)    # Get mask for relevant values

                    if self.heterogeneity_level == 'internal':

                        # Generate internal heterogeneity - hydraulic conductivity
                        k_flat = hu.specsim(self.grid,
                                            hydraulics['sig_y'][fi],
                                            hydraulics['ycorlengths'][fi],
                                            covmod='exp',
                                            selection_mask=mifi)

                        k_flat = np.exp(k_flat) * hydraulics['k_h'][fi]
                        k_iso[mifi] = k_flat

                        # Generate internal heterogeneity - porosity
                        n_flat = hu.specsim(self.grid, 
                                            hydraulics['sig_n'][fi],
                                            hydraulics['ncorlengths'][fi],
                                            covmod='exp',
                                            selection_mask=mifi) + hydraulics['n'][fi]
                        poros[mifi] = n_flat

                    elif self.heterogeneity_level == 'facies':
                        # Assign heterogeneity at facies level only
                        k_iso[mifi] = hydraulics['k_h'][fi]
                        poros[mifi] = hydraulics['n'][fi]

                    # Assign anisotropy ratio
                    anirat[mifi] = hydraulics['k_ratio'][fi]


            """ Assign background heterogeneity per architectural element """
            if self.heterogeneity_level == 'internal':
                # for si, aei in enumerate(ae_lu):
                for stratum in self.strata:
                    for ae in stratum.aes:
                        m0 = ha_arr == 0
                        ms = ae_arr == ae.num
                        ae_mask = m0 & ms     # Get material that equals zero within in architectural element
                        if np.all(ae_mask == False):
                            continue

                        ae_bg_facies = ae.bg_facies

                        # Assign background material
                        k_ae_flat = hu.specsim(self.grid,
                                            hydraulics['sig_y'][ae_bg_facies],
                                            hydraulics['ycorlengths'][ae_bg_facies], 
                                            covmod='exp',
                                            selection_mask=ae_mask)
                        k_ae_flat = np.exp(k_ae_flat) * hydraulics['k_h'][ae_bg_facies]          # back-transform from log space
                        k_iso[ae_mask] = k_ae_flat

                        # Generate internal heterogeneity - porosity
                        n_flat = hu.specsim(self.grid,
                                            hydraulics['sig_n'][ae_bg_facies],
                                            hydraulics['ncorlengths'][ae_bg_facies],
                                            covmod='exp',
                                            selection_mask=ae_mask)
                        n_flat = n_flat + hydraulics['n'][ae_bg_facies]
                        poros[ae_mask] = n_flat

                        # Assign background anisotropy ratio
                        anirat[ae_mask] = hydraulics['k_ratio'][ae_bg_facies]

                        # Assign architectural element trends
                        if ae.type_params['k_ztrend'] is not None:
                            # Vertical trend
                            zf_vec = np.linspace(*ae.type_params['k_ztrend'], self.gridnz)  # Z factor at each elevation
                            k_iso *= zf_vec
                        if ae.type_params['k_xtrend'] is not None:
                            xf_vec = np.linspace(*ae.type_params['k_xtrend'], self.grid.nx)
                            k_iso = np.transpose(k_iso.transpose() * xf_vec)

                """ Assign trends to hydraulic parameters globally """
                if hydraulics['k_ztrend'] is not None:
                    zf_vec = np.linspace(*hydraulics['k_ztrend'], self.grid.nz)  # Z factor at each elevation
                    k_iso *= zf_vec
                if hydraulics['k_xtrend'] is not None:
                    xf_vec = np.linspace(*hydraulics['k_xtrend'], self.grid.nx)
                    k_iso = np.transpose(k_iso.transpose() * xf_vec)

                if hydraulics['n_ztrend'] is not None:
                    zf_vec = np.linspace(*hydraulics['n_ztrend'], self.grid.nz)  # Z factor at each elevation
                    poros *= zf_vec
                if hydraulics['n_xtrend'] is not None:
                   xf_vec = np.linspace(*hydraulics['n_xtrend'], self.grid.nx)
                   poros = np.transpose(poros.transpose() * xf_vec)

        else:
            # Homogeneous case
            for hy_idx, hyi in enumerate(hydraulics['hydro']):
                hyi = hy_idx
                k_iso[fac == hyi] = hydraulics['k_h'][hy_idx]
                poros[fac == hyi] = hydraulics['n'][hy_idx]
                anirat[fac == hyi] = hydraulics['k_ratio'][hy_idx]

        # Assignment of anisotropy
        self.data['ktensors'] = np.zeros((self.grid.nx, self.grid.ny, self.grid.nz, 3, 3), dtype=np.float)
        ho.set_anisotropic_ktensor(self.data['ktensors'], k_iso, azim*np.pi/180, dip*np.pi/180, anirat)



    def print_model(self):
        """
        Prints a long description of the model. This is mainly for debugging.
        """

        print("Number of strata:", self.n_strata)
        print()
        print('====================================')
        print()
        for si in range(self.n_strata):
            print("Stratum:", self.strata[si].name)
            print("number of AEs:", self.strata[si].n_ae)
            print("zmin:", self.strata[si].zmin)
            print("zmax:", self.strata[si].zmax)
            print("---------------------------")
            for aei in range(self.strata[si].n_ae):
                print("AE:", self.strata[si].aes[aei].type_name)
                print("number of objects:", self.strata[si].aes[aei].n_objects)
                print("zmin", self.strata[si].aes[aei].zmin)
                print("zmax", self.strata[si].aes[aei].zmax)
                print("object_zmins:")
                print(self.strata[si].aes[aei].object_zmins)
                print("object_zmaxs:")
                print(self.strata[si].aes[aei].object_zmaxs)
                print()
            print()
            print('====================================')
            print()



# -*- coding: utf-8 -*-
""" Hydrogeological Virtual Reality simulation package.

    Hydrogeological virtual reality (HYVR) simulator for object-based modelling
    of sedimentary structures

    :Author: Jeremy P. Bennett

    :Notes:
         Grid nodes are cell-centred!


"""

import os
import sys
import numpy as np
import pickle
import math
import time
import scipy.io as sio
import matplotlib.pyplot as plt

import hyvr.utils as hu
import hyvr.optimized as ho
import hyvr.input.parameters as hp
from hyvr.postprocess.output import create_outputs
from hyvr.classes.model import Model
from hyvr.assign_facies import assign_facies_azim_dip


def run(param_file, flag_ow=None):
    """
    Main function for HYVR generation

    Parameters:
        param_file (str): 	Parameter file location
        flag_ow : bool
            If True, an existing run directory is overwritten.

    Returns:

        Save data outputs as parameter file

    """
    # TODO: implement a seed option to get the same model
    # np.random.seed(84)

    # Load parameter file
    run, model, hydraulics = hp.model_setup(param_file)
    hp.set_up_directories(run, param_file, flag_ow)

    for sim in range(1, int(run['numsim'])+1):
        # Generate facies
        model.generate_model()
        # model.assign_facies_azim_dip()
        assign_facies_azim_dip(model)
        # return

        if model.generate_hydraulics is True:
            # Generate internal heterogeneity
            model.generate_hydraulic_parameters(hydraulics)
        # from hyvr.postprocess.plotting import cross_section_pcolor
        # cross_section_pcolor(model, 'k_iso', log=True, y=10, xlim=[15.5, 16.4], ylim=[7.1, 7.7]
        # cross_section_pcolor(model, 'facies', log=False, y=10, xlim=[15.5, 16.4], ylim=[7.1, 7.7])
        # cross_section_pcolor(model, 'ha', log=False, y=10, xlim=[15.5, 16.4], ylim=[7.1, 7.7])
        # cross_section_pcolor(model, 'hat', log=False, y=10, xlim=[15.5, 16.4], ylim=[7.1, 7.7])
        # cross_section_pcolor(model, 'ae', log=False, y=10, xlim=[15.5, 16.4], ylim=[7.1, 7.7])
        # cross_section_pcolor(model, 'strata', log=False, y=10, xlim=[15.5, 16.4], ylim=[7.1, 7.7])
        # cross_section_pcolor(model, 'dip')
        # cross_section_pcolor(model, 'azim')

        # Save data
        if run['numsim'] > 1:
            realname = 'real_{:03d}'.format(sim)
            realdir = os.path.join(run['rundir'], realname)
        else:
            realname = run['runname']
            realdir = run['rundir']
        hu.try_makefolder(realdir)
        create_outputs(model, realdir, realname, run['outputs'])


    if param_file == 0:
        # this is just the testcase, so we remove the output
        from pkg_resources import cleanup_resources
        cleanup_resources()
        import shutil
        runpath = os.path.abspath(run['rundir'])
        if os.path.exists(runpath):
            shutil.rmtree(runpath)



import numpy as np

def parse_contact_model(param_list, depth=False):
    mode = param_list[0].strip("\"'")
    cm = {'mode': mode}
    param_idx = 1
    if depth:
        if len(param_list) < 2:
            raise ValueError("This is not a valid contact model: " + str(param_list))
        param_idx = 2
        cm['z'] = float(param_list[1])

    # make sure the length is right
    if mode.lower() == 'flat' and len(param_list) - param_idx == 0:
        return cm
    elif mode.lower() == 'random' and len(param_list) - param_idx == 3:
        cm['var'] = float(param_list[param_idx])
        cm['corlx'] = float(param_list[param_idx+1])
        cm['corly'] = float(param_list[param_idx+2])
        return cm
    else:
        raise ValueError("This is not a valid contact model: " + str(param_list))


def use_lower_surface_value(target_surface, other_surface):
    nx, ny = np.shape(target_surface.surface)
    for i in range(nx):
        for j in range(ny):
            if target_surface.surface[i,j] > other_surface.surface[i,j]:
                target_surface.surface[i,j] = other_surface.surface[i,j]
    target_surface.zmax = np.max(target_surface.surface)

def use_higher_surface_value(target_surface, other_surface):
    nx, ny = np.shape(target_surface.surface)
    for i in range(nx):
        for j in range(ny):
            if target_surface.surface[i,j] < other_surface.surface[i,j]:
                target_surface.surface[i,j] = other_surface.surface[i,j]
    target_surface.zmax = np.max(target_surface.surface)




"""
This is the main module for parsing the parameter file. There are two other
modules strongly linked ot this one:

* options: contains definitions of the possible options of a parameter file
* option_parsing: provides the classes Section and Option for simpler parsing and validation.

:Authors: Jeremy P. Bennett, Samuel Scherrer
"""


import configparser as cp
import os
import sys
import shutil
from hyvr.utils import try_makefolder, print_to_stdout
from hyvr.input.options import options
from hyvr.input.options_deprecated import options as old_options
from hyvr.input.option_parsing import *
from hyvr.classes.model import Model
from hyvr.classes.contact_surface_utils import parse_contact_model


def model_setup(pf):
    """
    Set up model using grid.Grid() class and assign parameters

    Parameters:
        pf (str):   					Parameter file path

    Returns:
        - run *(dict)* - Model run parameters
        - mod *(dict)* - Model domain parameters
        - sequences *(dict)* - Sequence parameters
        - hydraulics *(dict)* - Hydraulic properties parameters
        - flowtrans *(dict)* - Flow & transport simulation parameters
        - elements *(dict)* - Architectural elements and parameters
        - model_grid *(object class)* - Grid object class

    """
    run, model_dict, strata_dict, hydraulics, flowtrans, elements = parameters(pf)
    # Assign architectural element identifiers
    for element in elements.keys():
        elements[element]['ae_id'] = strata_dict['ae'].index(element)

    # create model object
    model = Model(model_dict, strata_dict, elements, flowtrans)
    return run, model, hydraulics


def parameters(inifile):
    """
    Get parameters for hierarchical facies modelling from a .ini-file

    Parameters:
        inifile (str):  		Parameter file path

    Returns:
        - run *(dict)* - Model run parameters
        - model *(dict)* - Model domain parameters
        - sequences *(dict)* - Sequence parameters
        - hydraulics *(dict)* - Hydraulic properties parameters
        - flowtrans *(dict)* - Flow & transport simulation parameters
        - elements *(dict)* - Architectural elements and parameters
    """

    print_to_stdout("Reading parameter file")

    # test case
    if inifile == 0:
        from pkg_resources import resource_filename
        inifile = resource_filename(__name__, os.path.join(os.pardir, 'made.ini'))

    # read file
    p = cp.ConfigParser()
    try:
        p.read(inifile, encoding='utf-8')
    except cp.MissingSectionHeaderError:
        # this is probably caused by a wrong encoding
        p.read(inifile, encoding='utf-8-sig')
    if len(p.sections()) == 0:
        raise FileNotFoundError("Parameter file {:s} not found!".format(inifile))

    old_format = 'dataoutputs' in dict(p['run'])
    if old_format:
        run, model, strata, hydraulics, flowtrans, elements = get_new_parameters_from_deprecated(*parse_deprecated_inifile(p))


    else:
        run, model, strata, hydraulics, flowtrans, elements = parse_inifile(p)

    # Runname and Modeldir
    # ====================

    # if runname is not given, it's the part of the ini-file before
    # "_parameters.ini" or ".ini"
    # in the test case runname is given
    ini_path = os.path.dirname(os.path.abspath(inifile))
    ini_name = os.path.basename(os.path.abspath(inifile))
    if run['runname'] is None:
        if ini_name[-25:] == "autogenerated_backup.ini":
            run['runname'] = ini_name[0:-25]
        else:
            run['runname'] = ini_name[0:-4]


    # separate task: find modeldir

    # modeldir is either the given option or the path of the .ini file
    # rundir is modeldir/runname
    # But:
    # If the .ini-filename ends with "autogenerated_backup.ini", the directory of the
    # ini-file is the rundir and modeldir is the directory above.
    # This should overwrite all other settings (except overwrite_old_output)
    if ini_name[-25:] == "_autogenerated_backup.ini":
        run['modeldir'] = os.path.abspath(os.path.join(ini_path, os.pardir))
        run['rundir'] = os.path.abspath(ini_path)
    else:
        # either inipath or the chosen modeldir
        if run['modeldir'] is None:
            run['modeldir'] = ini_path
        if run['modeldir'] == 'select':
            run['modeldir'] = input(
                'Please input the model directory save path, or press <enter> to'
                'save in default directory:\n')
            if len(run['modeldir']) == 0:
                run['modeldir'] = inipath
        run['modeldir'] = os.path.abspath(run['modeldir'])
        run['rundir'] = os.path.join(run['modeldir'], run['runname'])


    return run, model, strata, hydraulics, flowtrans, elements


def parse_inifile(p):

    # TODO: these names were at some time keywords for hyvr, but I don't know what they describe and
    # they are not implemented anymore
    str_values = 'k_trend', 'linear_acceleration'

    sections = p.sections()
    section_parser = {}

    # The following code is not very nice, with lots of repetitions. This could
    # be much nicer if the strata section had only one possible name.
    must_haves = ['run', 'model', 'strata', 'hydraulics']
    for section in must_haves:
        if section not in sections:
            raise MissingSectionError(section)

    # run section
    # -----------
    run = Section('run', options['run']).parse(dict(p['run']))
    del sections[sections.index('run')]

    # hydraulics section
    # ------------------
    hydraulics = Section('hydraulics', options['hydraulics']).parse(dict(p['hydraulics']))
    del sections[sections.index('hydraulics')]

    # model section
    # -------------
    model = Section('model', options['model']).parse(dict(p['model']))
    del sections[sections.index('model')]
    if model['dy'] == None:
        model['dy'] = model['dx']
    if model['dz'] == None:
        model['dz'] = model['dx']

    # strata section
    # --------------
    strata = Section('strata', options['strata']).parse(dict(p['strata']))
    del sections[sections.index('strata')]
    if len(strata['strata_contact_models']) != len(strata['strata']) - 1:
        raise ShapeError('strata_contact_models', 'strata')
    strata['contact_models'] = [
        parse_contact_model(model, depth=True) for model in strata['strata_contact_models']
    ]
    try:
        strata['bg_facies'] = parse_facies(strata['bg_facies'], hydraulics['hydrofacies'])
    except:
        raise ValueError('Invalid facies string in strata section in option bg_facies: ' + strata['bg_facies'])


    # flowtrans section
    # ------------------
    # this section is only necessary for model output
    for output in ['mf', 'mf6', 'hgs']:
        if output in run['outputs']:
            if 'flowtrans' not in sections:
                raise MissingSectionError('flowtrans')
            break
    if 'flowtrans' in sections:
        flowtrans = Section('flowtrans', options['flowtrans']).parse(dict(p['flowtrans']))
        del sections[sections.index('flowtrans')]
    else:
        flowtrans = {}

    # remaining sections are architectural elements
    elements = {}
    for section in sections:
        dictionary = dict(p[section])
        assert_exists('geometry', dictionary, section)
        geometry = dictionary['geometry']
        if geometry not in ['trough', 'channel', 'sheet']:
            raise ValueError('Invalid geometry: ' + geometry + ' in section ' + section)
        elements[section] = Section(section, options[geometry]).parse(dict(p[section]))
        elements[section]['contact_model'] = parse_contact_model(elements[section]['contact_model'], depth=False)
        # get facies number
        try:
            elements[section]['facies'] = parse_facies(elements[section]['facies'], hydraulics['hydrofacies'])
        except:
            raise ValueError('Invalid facies string in section ' + section + ' in option facies!')
        # get altfacies
        if elements[section]['altfacies'] is not None:
            try:
                elements[section]['altfacies'] = [
                    parse_facies(facies_list, hydraulics['hydrofacies']) for facies_list in elements[section]['altfacies']
                ]
            except:
                raise ValueError('Invalid facies string in section ' + section + ' in option altfacies!')
        # get bg_facies number
        if elements[section]['bg_facies'] == None:
            elements[section]['bg_facies'] = -1
        else:
            try:
                elements[section]['bg_facies'] = hydraulics['hydrofacies'].index(elements[section]['bg_facies'])
            except:
                raise ValueError('Invalid facies in section ' + section + ' in option bg_facies: ', elements[section]['bg_facies'])


        # if structure is 'dip', require 'dipset_dist'
        if elements[section]['structure'] == 'dip' or elements[section]['structure'] == 'random':
            assert_exists('dipset_dist', dictionary, section)
        # if structure is 'bulb_sets', require 'bulbset_dist'
        if elements[section]['structure'] == 'bulb' or elements[section]['structure'] == 'random':
            assert_exists('bulbset_dist', dictionary, section)

        # lag surface facies
        if geometry in ['trough', 'channel']:
            if elements[section]['lag_height'] != 0.0:
                try:
                    elements[section]['lag_facies'] = hydraulics['hydrofacies'].index(elements[section]['lag_facies'])
                except:
                    raise ValueError('lag_facies in section ' + section + ' is invalid!')
            else:
                elements[section]['lag_facies'] = -1



    return run, model, strata, hydraulics, flowtrans, elements

def get_new_parameters_from_deprecated(run, model, strata, hydraulics, flowtrans, elements):
    """
    This applies the necessary changes to parsed sections of an old-format inifile.
    """

    # run section
    run['outputs'] = run['modeloutputs'] + run['dataoutputs']
    run['overwrite_old_output'] = run['flag_ow']

    # hydraulics section
    hydraulics['hydrofacies'] = hydraulics['hydro']


    # model section
    model['anisotropy'] = run['anisotropy']
    model['hydraulics'] = hydraulics['gen']
    model['heterogeneity'] = run['het']
    model['heterogeneity_level'] = model['hetlev']
    if model['display']:
        raise ValueError('display option is not supported anymore')


    # strata section
    n_strata = len(strata['ssm'])
    strata['bg_facies'] = [int(strata['bg'][0])] * n_strata
    strata['bg_azim'] = [strata['bg'][1]] * n_strata
    strata['bg_dip'] = [strata['bg'][2]] * n_strata
    strata['strata'] = strata['ssm']
    if strata['ae_table'] is not None:
        raise ValueError('AE tables are not supported anymore!')
    if strata['save_aelu']:
        raise ValueError('save_aelu is not supported anymore!')
    contact_models = []
    for i in range(n_strata-1):
        cm = parse_old_strata_contact_model(strata, i)
        contact_models.append(cm)
    strata['contact_models'] = contact_models
    strata['ae_in_strata'] = strata['ssm_ae']

    # element sections
    for elem_name in elements:
        elem = elements[elem_name]
        elem['size_ztrend'] = elem['geo_ztrend']
        if elem['bg'] is not None:
            elem['bg_facies'] = elem['bg'][0]
            elem['bg_azim'] = elem['bg'][1]
            elem['bg_dip'] = elem['bg'][1]
        else:
            elem['bg_facies'] = -1
            elem['bg_azim'] = float('NaN')
            elem['bg_dip'] = float('NaN')
        elem['dipset_dist'] = elem['dipset_d']
        if 'bulbset_d' in elem:
            elem['bulbset_dist'] = elem['bulbset_d']
        if elem['geometry'] == 'trunc_ellip':
            elem['geometry'] = 'trough'
            elem['trough_density'] = elem['el_z']
        elem['contact_model'] = parse_old_elem_contact_model(elem)
        if elem['geometry'] in ['trough', 'channel']:
            if elem['lag'] is not None:
                elem['lag_height'] = elem['lag'][0]
                elem['lag_facies'] = elem['lag'][1]
            else:
                elem['lag_height'] = 0.0
                elem['lag_facies'] = -1



    return run, model, strata, hydraulics, flowtrans, elements


def parse_facies(facies_list, hydrofacies):
    """
    Reads a list of facies strings (facies_list) and returns the corresponding
    indices in the hydrofacies list.
    """
    return [
        hydrofacies.index(facies) for facies in facies_list
    ]

def get_new_facies_list(facies_list, hydrofacies):
    return [hydrofacies[i] for i in facies_list]


def parse_old_strata_contact_model(strata, i):
    mode = strata['ssm_contact']
    cm = {'mode':mode, 'z':strata['ssm_top'][i]}
    if mode.lower() == 'flat':
        return cm
    elif mode.lower() == 'random':
        cm['var'] = strata['ssm_contact_model'][i][0]
        cm['corlx'] = strata['ssm_contact_model'][i][1]
        cm['corly'] = strata['ssm_contact_model'][i][2]
        return cm
    else:
        raise ValueError('Unknown contact type in strata section: ' + mode)

def parse_old_elem_contact_model(elem):
    mode = elem['contact']
    cm = {'mode':mode}
    if mode.lower() == 'flat':
        return cm
    elif mode.lower() == 'random':
        cm['var'] = elem['contact_model'][0]
        cm['corlx'] = elem['contact_model'][1]
        cm['corly'] = elem['contact_model'][2]
        return cm
    else:
        raise ValueError('Unknown contact type in strata section: ' + mode)



def parse_deprecated_inifile(p):

    import warnings
    print()
    warnings.warn("You seem to be using the old ini-file format. We strongly recommend to use the new format as described in our documentation.", DeprecationWarning)
    print("------- Warning: You are using the old ini-file format ----------------------")
    print("We strongly recommend to use the new format as described in the documentation")
    print("-----------------------------------------------------------------------------")
    print()

    sections = p.sections()
    section_parser = {}

    # The following code is not very nice, with lots of repetitions. This could
    # be much nicer if the strata section had only one possible name.
    must_haves = ['run', 'model', 'strata', 'hydraulics', 'flowtrans']
    for section in must_haves:
        if section not in sections:
            raise MissingSectionError(section)

    run = Section('run', old_options['run']).parse(dict(p['run']))
    del sections[sections.index('run')]

    model = Section('model', old_options['model']).parse(dict(p['model']))
    del sections[sections.index('model')]
    if model['dy'] == None:
        model['dy'] = model['dx']
    if model['dz'] == None:
        model['dz'] = model['dx']

    strata = Section('strata', old_options['strata']).parse(dict(p['strata']))
    del sections[sections.index('strata')]
    if strata['ae_table'] is not None:
        if not os.path.exists(os.path.join(os.path.dirname(inifile), strata['ae_table'])):
            raise FileNotFoundError('ae_table-file {:s} not found!'.format(strata['ae_table']))

    hydraulics = Section('hydraulics', old_options['hydraulics']).parse(dict(p['hydraulics']))
    del sections[sections.index('hydraulics')]

    flowtrans = Section('flowtrans', old_options['flowtrans']).parse(dict(p['flowtrans']))
    del sections[sections.index('flowtrans')]

    # remaining sections are architectural elements
    elements = {}
    for section in sections:
        dictionary = dict(p[section])
        assert_exists('geometry', dictionary, section)
        geometry = dictionary['geometry']
        elements[section] = Section(section, old_options[geometry]).parse(dict(p[section]))


    return run, model, strata, hydraulics, flowtrans, elements

def set_up_directories(run, configfile, overwrite_old_output=None):
    """
    This functions creates the necessary directories (modeldir, rundir). It also stores the used
    ini-file in the rundir.

    Parameters:
        run (dict):        parsed run-section of the config file
        configfile (str):  path to config file
        overwrite_old_output (bool):    Whether to overwrite the old run directory. If it is None, the option
                           from the config file will be chosen instead (default in configfile: False)
    """

    # test case
    if configfile == 0:
        from pkg_resources import resource_filename
        configfile = resource_filename(__name__, 'made.ini')

    p = cp.ConfigParser()
    try:
        p.read(configfile, encoding='utf-8')
    except cp.MissingSectionHeaderError:
        # this is probably caused by a wrong encoding
        p.read(configfile, encoding='utf-8-sig')

    # we're now done with parsing and can create the model directory.
    try_makefolder(run['modeldir'])

    # If the run directory already exists, it is overwritten if overwrite_old_output is set to True.
    # If it doesn't exist, it will be created.
    if overwrite_old_output is None:
        overwrite_old_output = run['overwrite_old_output']
    if os.path.exists(run['rundir']):
        if run['overwrite_old_output'] is True:
            # If it exists, we just delete everything in it
            for f in os.listdir(run['rundir']):
                path = os.path.join(run['rundir'], f)
                if os.path.isfile(path):
                    os.remove(path)
                elif os.path.isdir(path):
                    shutil.rmtree(path)
        else:
            raise FileExistsError("Run directory already exists and overwrite flag is "
                            "set to 'false'" ". Either change the runname, or "
                            "change the overwrite flag ('overwrite_old_output') in the config file or run hyvr "
                            "with --overwrite.")
    else:
        os.makedirs(run['rundir'])

    file_save = os.path.join(run['rundir'], run['runname'] + '_autogenerated_backup.ini')
    with open(file_save, 'w') as configfile:
        p.write(configfile)



# -*- coding: utf-8 -*-
"""Some utility functions for HFM modelling

    :Authors: Jeremy P. Bennett, with help from Alessandro Comunian and Samuel Scherrer

    :Notes:



"""

import sys
import pickle
import numpy as np
import time
import configparser as cp
import pandas as pd
import linecache
import scipy.io
import scipy.spatial as sp
import os
import shutil
import errno
import scipy.stats as st
import hyvr.optimized as ho




def mf6_vtr(fhead, mg, fout):
    """
    Convert a MODFLOW 6 Binary head file into vtr suitable for visualisation in ParaView

    Parameters
    ----------



    """
    try:
        import flopy
    except ImportError:
        print('mf output not possible: Flopy not installed.')
        return
    hfile = flopy.utils.binaryfile.HeadFile(fhead)          #
    hdata = hfile.get_alldata()                             # Create numpy array with all data
    head_dict = dict()                                      # Initialise dict with the data
    for i in range(0, hdata.shape[0]):
        hf_i = np.squeeze(hdata[i, :, :, :])                # Get heads at individual time steps
        hf_i = np.transpose(hf_i, (2, 1, 0))                # Permute to be consistent with HyVR grids
        head_dict['head_timestep{}'.format(i)] = hf_i

    to_vtr(head_dict, fout, mg, points=True)


def dem_load(fn):
    """
    Load data from ESRI-style ASCII-file.

    Parameters:
        fn (str): 				Directory and file name for save

    Returns:
        - data *(numpy array)* - Data from ERSI-style ASCII-file
        - meta *(dict)* - Dict with grid metadata

    """

    # Extract header using linecache
    meta = {}
    meta['ncols'] = int(linecache.getline(fn, 1).split()[1])
    meta['nrows'] = int(linecache.getline(fn, 2).split()[1])
    meta['ox'] = linecache.getline(fn, 3).split()[1]
    meta['oy'] = linecache.getline(fn, 4).split()[1]
    meta['cell_size'] = linecache.getline(fn, 5).split()[1]
    meta['no_Data'] = linecache.getline(fn, 6).split()[1]

    # Extract data using pandas
    df = pd.read_csv(fn, header=None, delimiter=' ', skiprows=6, dtype=np.float)
    data = df.as_matrix()

    return data, meta


def dem_save(fn, data, gro):
    """
    Save DEM data to ESRI-style ASCII-file

    Parameters:
        fn (str):               Directory and file name for save
        data (numpy array):     DEM data
        gr (object class):      grid.Grid() object class

    Returns:
        Save DEM data to ESRI-style ASCII-file

    """

    header = ("ncols            {0.nx}\n"
              "nrows            {0.ny}\n"
              "xllcorner        {0.ox}\n"
              "yllcorner        {0.oy}\n"
              "cellsize         {0.cs2}\n"
              "NOODATA_value    -9999"
              ).format(gro)

    with open(fn, mode='wb') as out_file:
        np.savetxt(out_file,
                   data,
                   header=header,
                   fmt='%.4f',
                   comments='')



def load_gslib(fn):
    """
    Load .gslib files. This has been appropriated from the HPGL library
    https://github.com/hpgl/hpgl/blob/master/src/geo_bsd/routines.py
    commit b980e15ad9b1f7107fd4fa56ab117f45553be3aa

    Parameters:
        fn (str): 			.gslib file path and name

    Returns:
        gslib_dict *(dict)* - properties

    """
    gslib_dict = {}
    list_prop = []
    points = []

    f = open(fn)
    head = f.readline().split('\t')
    num_p = int(f.readline())
    #print num_p

    lx, ly, lz = [int(x) for x in head[0].split(' ')]
    nx, ny, nz = [float(x) for x in head[1].split(' ')]
    ox, oy, oz = [float(x) for x in head[2].split(' ')]

    for i in range(num_p):
        list_prop.append(str(f.readline().strip()))
    #print list_prop

    for i in range(len(list_prop)):
        gslib_dict[list_prop[i]] = np.zeros((lx * ly * lz))

    index = np.zeros(len(list_prop))

    for line in f:
        points = line.split()
        for j in range(len(points)):
            gslib_dict[list_prop[j]][index[j]] = float(points[j])
            index[j] += 1

    for dkey in gslib_dict.keys():
        gslib_dict[dkey] = gslib_dict[dkey].reshape((ly, lx, lz))

    f.close()

    return gslib_dict


''' HYVR-specific utilities'''


def read_lu(sq_fp):
    """
    Load user-defined strata (architectural element lookup table),
    split the data based on a delimiter and return it as a new list

    Parameters:
        sq_fp:			Load user-defined strata (architectural element lookup table)

    Returns:
        ssm_lu *(list)*: -Values of architectural element lookup table

    """
    # Load user-defined systems / architectural element lookup table
    print(time.strftime("%d-%m %H:%M:%S", time.localtime(time.time())) + ': Reading strata data from ' + sq_fp)
    with open(sq_fp) as f:
        lines = f.read().splitlines()

    ssm_lu = []
    for li in lines[1:]:
        temp = li.split(',')
        ssm_lu.append([int(temp[0]), float(temp[1]), float(temp[2]), str(temp[3]), int(temp[4])])

    return ssm_lu



def virtual_boreholes(data_dict, d, l, file_out=None, vals=[], opts=[]):
    """ Perform 'virtual' borehole sampling of parameter field

    Arguments:
        data_dict (dict):
            Data to sample
        d (list):
            3-tuple of model grid cell dimensions
        l (list):
            3-tuple of total model dimensions/lengths
        file_out (list: basefile path, list of file types):
            Output filename and path
        vals (list):
            Parameter fields to include
        opts (dict):
            Sampling options
            opts['noBH'] (int):             Random sampling
            opts['grid_spacing']:           float, or list of floats [x spacing, y spacing] Grid sample spacing
            opts['grid_n']:                 int, or list of ints [n in x, n in y] Number of grid nodes per x,y dimensions
            opt['lnK'] (bool):              Natural logarithm of isotropic hydraulic conductivity

    Returns:
        bh_df : Pandas DataFrame class

    """

    nx, ny, nz = np.shape(data_dict['fac'])

    # Set up column names
    cols = ['x', 'y', 'z']
    if len(vals) == 0:
        vals = data_dict.keys()
    cols.extend(vals)

    # Create dataframe
    bh_df = pd.DataFrame(columns=cols)

    # Sampling of grid
    xy_grid = []
    xv = np.arange(0.5 * d[0], l[0], d[0])
    yv = np.arange(0.5 * d[1], l[1], d[1])
    zv = np.arange(0.5 * d[2], l[2], d[2])

    if 'grid_spacing' in opts.keys():
        """ Uniform grid with set spacing """

        # Get cartesian coordinates in 2D (x,y)
        if len(opts['grid_spacing']) == 2:
            range_x = np.arange((opts['grid_spacing'][0] * 0.5), l[0], opts['grid_spacing'][0])
            range_y = np.arange((opts['grid_spacing'][1] * 0.5), l[1], opts['grid_spacing'][1])
        else:
            range_x = np.arange((opts['grid_spacing'] * 0.5), l[0], opts['grid_spacing'])
            range_y = np.arange((opts['grid_spacing'] * 0.5), l[1], opts['grid_spacing'])

        x_locs, y_locs = np.meshgrid(range_x, range_y)

        # Convert to array indices
        x_locs = np.floor(x_locs.flatten()/d[0]).astype(int)
        y_locs = np.floor(y_locs.flatten()/d[1]).astype(int)

    elif 'grid_n' in opts.keys():
        """ Sample over uniform grid """
        # Get cartesian coordinates in 2D (x,y)
        range_x = np.linspace(0, l[0]/d[0] - 1, opts['grid_n'][0])
        range_y = np.linspace(0, l[1]/d[1] - 1, opts['grid_n'][1]+2)[1:-1]
        x_locs, y_locs = np.meshgrid(range_x, range_y)

        # Convert to array indices
        x_locs = np.floor(x_locs.flatten()).astype(int)
        y_locs = np.floor(y_locs.flatten()).astype(int)

    elif 'noBH' in opts.keys():
        """ Randomly sample the xy plane """
        x_locs = np.random.choice(range(0, nx), opts['noBH'])    # Borehole location indices
        y_locs = np.random.choice(range(0, ny), opts['noBH'])    # Borehole location indices

    # Put data into dataframe
    for idx in range(len(x_locs)):
        # Get indices of location
        i = x_locs[idx]
        j = y_locs[idx]

        # Get vectors of Cartesian coordinates
        ibh = np.zeros((nz, 3 + len(vals)))
        ibh[:, 0] = np.ones((nz,)) * xv[i]          # x coordinates
        ibh[:, 1] = np.ones((nz,)) * yv[j]          # y coordinates
        ibh[:, 2] = zv                              # z coordinates

        for iv, v in enumerate(vals):
            # Append to list to be appended to dataframe
            ibh[:, iv+3] = data_dict[v][i, j, 0:nz]
        bh_df = bh_df.append(pd.DataFrame(ibh, columns=cols), ignore_index=True)

    if 'lnK' in opts and opts['lnK'] is True:
        vals.extend(['lnK'])
        bh_df['lnK'] = pd.Series(np.log(bh_df['k_iso']), index=bh_df.index)

    # Save borehole data
    fmtd = {'k_iso': '%.5e',
            'lnK': '%.5f',
            'poros': '%.5f',
            'fac': '%u',
            'dip': '%.2f',
            'azim': '%.2f'}
    if file_out is not None:
        if 'csv' in file_out[1]:
            bh_df.to_csv(file_out[0]+'.csv', index=False)
        if 'gslib' in file_out[1]:
            n_conddata = bh_df.shape[0]
            colsout = ['x', 'y', 'z']
            colsout.extend(vals)
            to_write = bh_df.as_matrix(columns=colsout)
            header = [str(n_conddata), str(3 + len(vals)), 'x', 'y', 'z']
            header.extend(vals)
            header = '\n'.join(header)

            fmts = '%.3f %.3f %.3f {}'.format(' '.join([fmtd[i] for i in vals]))
            np.savetxt(file_out[0] + '.gslib', to_write, delimiter=' ', header=header, comments='', fmt=fmts)

    return bh_df


def calc_norm(x):
    """
    Calculate norm (compute the complex conjugate from 'x')

    Parameters:
        x:	Input parameter

    Returns:
        Complex conjugate of x

    """
    s = (x.conj() * x).real
    return np.sqrt(np.add.reduce(s, axis=0))


def try_makefolder(makedir):
    """
    Create modflow output folder

    """
    # Create modflow output folder

    try:
        os.makedirs(makedir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def print_to_stdout(*args):
    """
    Prints a message to stdout with timestamp.
    """
    print(time.strftime("%d-%m %H:%M:%S", time.localtime(time.time())) + ':', *args)


####################################################################################################################
# Some utilities for generating shapes
####################################################################################################################

def planepoint(dip_norm, x_dip, y_dip, znow, xtemp, ytemp, ztemp, select=[]):
    """
    Compute number of planes

    Parameters:
        dip_norm:
        x_dip:			X coordinates of points on dip planes
        y_dip:			Y coordinates of points on dip planes
        znow:			Current coordinates of Z, needed to compute Z coordinates of points on dip planes
        xtemp:			X dimension of model grid nodes
        ytemp:			Y dimension of model grid nodes
        ztemp:			Z dimension of model grid nodes
        select:         Model grid nodes to consider

    Returns:
        set_no - Number of planes with selected model grid nodes

    """
    # Get closest plane to points
    n_sets = dip_norm.shape[1]                   # Number of planes
    nx, ny, nz = xtemp.shape                       # Get number of model cells
    set_no = np.zeros((nx, ny, nz), dtype=np.int)  # Initialise set number array
    z_dip = np.ones(x_dip.shape) * znow

    if len(select) > 0:
        select_idx = np.where(select)                                               # Get indices of selected model nodes
    else:
        select = np.ones_like(xtemp, dtype=bool)
        select_idx = np.where(select)

    points = np.array((xtemp[select].flatten(), ytemp[select].flatten(), ztemp[select].flatten()))      # Cartesian coordinates of model grid nodes
    plp = np.array((x_dip, y_dip, z_dip)).T                                     # Cartesian coordinates of points on dip planes
    pd = plp[:, None] - points.T                                                # subtract grid nodes from plane points

    # Loop over set planes
    for iset in range(n_sets-1):
        if iset > 1:
            pd_1 = pd_2
        else:
            abc_1 = dip_norm[:, iset]                                                           # Plane normal equation
            pd_1 = abc_1.dot(pd[iset, :, :].squeeze().T) / np.sqrt(sum(abc_1 * abc_1))          # Distance to plane
        pd1_c1 = pd_1 <= 0                                                                  # pd_1 meeting condition 1
        pd1_c1_idx = np.where(pd1_c1)

        if iset == 0:
            set_no[select_idx[0][pd1_c1_idx], select_idx[1][pd1_c1_idx], select_idx[2][pd1_c1_idx]] = iset+1
        # elif iset == n_sets: # this never happens
        #     pd1_c2_idx = np.where(pd_1 > 0)                     # index of pd_2 meeting condition 1
        #     set_no[select_idx[0][pd1_c2_idx], select_idx[1][pd1_c2_idx], select_idx[2][pd1_c2_idx]] = iset+1
        else:
            abc_2 = dip_norm[:, iset+1]
            # Points on plane
            pd_2 = abc_2.dot(pd[iset+1, :, :].squeeze().T) / np.sqrt(sum(abc_2 * abc_2))  # Distance to plane
            inset = np.logical_and(pd_1 <= 0, pd_2 > 0)                                   # grid cell between planes
            set_no[select_idx[0][inset], select_idx[1][inset], select_idx[2][inset]] = iset+1

    return set_no


def angle(v1, v2):
    """
    Return angle between two vectors in [°] between 0° and 180°

    Parameters:
        v1:	Vector 1
        v2:	Vector 2

    Returns:
        angle value *(float)* - Angle between v1 and v2
    """
    cos = np.dot(v1, v2) / np.sqrt(np.dot(v1, v1) * np.dot(v2, v2))
    angle = np.arccos(cos)/np.pi*180
    return angle

def get_alternating_facies(num_facies, type_params):
    """
    Returns a vector of alternating facies numbers based on the 'altfacies'
    setting in the inifile.
    """
    facies = np.zeros(num_facies, dtype=np.int32)
    facies[0] = np.random.choice(type_params['facies'])
    # The facies are changing according to the given 'altfacies'
    if type_params['altfacies'] is not None:
        for i in range(1, num_facies):
            fac_idx = type_params['facies'].index(facies[i-1])
            facies[i] = np.random.choice(type_params['altfacies'][fac_idx])
    else:
        for i in range(1, num_facies):
            facies[i] = np.random.choice(type_params['facies'])
    return facies

def norm(v):
    return np.sqrt(np.dot(v,v))


def specsim(grid, var, corl, selection_mask=None, two_dim=False, covmod='gaussian'):
    """
    Generate random variables with stationary covariance function using spectral
    techniques of Dietrich & Newsam (1993)

    Parameters
    ----------
    grid : Grid instance
    var : float
        Variance
    corl : tuple of floats
        Tuple of correlation lengths. 2-tuple for 2-d (x and y), and 3-tuple
        for 3-d (x, y, z).
    selection_mask : boolean numpy array
        This array should have the same size as the model grid and can be used to only generate
        random variables for a subset of the model grid. If ``two_dim=True``, it should have
        dimensions ``(nx, ny)``, otherwise ``(nx, ny, nz)``.
    two_dim : bool, optional (default: False)
        Whether to return a two-dimensional or a 3 dimensional field
    covmod : str, optional (default: "gaussian")
        Which covariance model to use ("gaussian" or "exp").

    Returns
    -------
    Y : 1d, 2d, or 3d numpy array
        Numpy array of random field. If no selection mask was given, this is either a 2d or 3d numpy
        array, depending on ``two_dim`` and of the size of the model grid.
        If a selection_mask was given, this is a flat array of values.
    """

    # THINK: If this is called multiple times in a row with the same parameters (i.e. same var,
    # corl, covmod), then it is not necessary to recalculate syy, and we even get two fields out.
    # Is there a way to use this?
    mask_given = selection_mask is not None
    if mask_given:
        # This method only works for rectangular grids, this means we have to find the smallest
        # rectangular selection that contains all selected cells
        if two_dim:
            x_idx, y_idx = np.where(selection_mask)
            min_x_idx = min(x_idx)
            max_x_idx = max(x_idx)
            min_y_idx = min(y_idx)
            max_y_idx = max(y_idx)
            rectangular_mask = np.zeros_like(selection_mask)
            rectangular_mask[min_x_idx:max_x_idx+1,min_y_idx:max_y_idx+1] = True
            selected_X = grid.X[min_x_idx:max_x_idx+1,min_y_idx:max_y_idx+1,0]
            selected_Y = grid.Y[min_x_idx:max_x_idx+1,min_y_idx:max_y_idx+1,0]
            selected_X_centered = selected_X - (np.min(selected_X) + np.max(selected_X))/2
            selected_Y_centered = selected_Y - (np.min(selected_Y) + np.max(selected_Y))/2
            selection_mask_small = selection_mask[min_x_idx:max_x_idx+1,min_y_idx:max_y_idx+1]

            h_square = (selected_X_centered/corl[0])**2 \
                    + (selected_Y_centered/corl[1])**2

        else:
            x_idx, y_idx, z_idx = np.where(selection_mask)
            try:
                min_x_idx = min(x_idx)
            except:
                import pdb; pdb.set_trace()
            max_x_idx = max(x_idx)
            min_y_idx = min(y_idx)
            max_y_idx = max(y_idx)
            max_z_idx = max(z_idx)
            min_z_idx = min(z_idx)
            rectangular_mask = np.zeros_like(selection_mask)
            rectangular_mask[min_x_idx:max_x_idx+1,min_y_idx:max_y_idx+1,min_z_idx:max_z_idx+1] = True
            selected_X = grid.X[min_x_idx:max_x_idx+1,min_y_idx:max_y_idx+1,min_z_idx:max_z_idx+1]
            selected_Y = grid.Y[min_x_idx:max_x_idx+1,min_y_idx:max_y_idx+1,min_z_idx:max_z_idx+1]
            selected_Z = grid.Z[min_x_idx:max_x_idx+1,min_y_idx:max_y_idx+1,min_z_idx:max_z_idx+1]
            selected_X_centered = selected_X - (np.min(selected_X) + np.max(selected_X))/2
            selected_Y_centered = selected_Y - (np.min(selected_Y) + np.max(selected_Y))/2
            selected_Z_centered = selected_Z - (np.min(selected_Z) + np.max(selected_Z))/2
            selection_mask_small = selection_mask[min_x_idx:max_x_idx+1,min_y_idx:max_y_idx+1,min_z_idx:max_z_idx+1]

            h_square = (selected_X_centered/corl[0])**2 \
                     + (selected_Y_centered/corl[1])**2 \
                     + (selected_Z_centered/corl[2])**2

    else:
        # full grid
        if two_dim:
            h_square = (np.asarray(grid.X_centered[:,:,0])/corl[0])**2 \
                    + (np.asarray(grid.Y_centered[:,:,0])/corl[1])**2
        else:
            h_square = (np.asarray(grid.X_centered)/corl[0])**2 \
                     + (np.asarray(grid.Y_centered)/corl[1])**2 \
                     + (np.asarray(grid.Z_centered)/corl[2])**2

    ntot = h_square.size

    # Covariance matrix of variables
    if covmod == 'gaussian':
        # Gaussian covariance model
        ryy = np.exp(-h_square) * var
    elif covmod == 'exp':
        # Exponential covariance model
        ryy = np.exp(-np.sqrt(h_square)) * var
    else:
        raise ValueError('Invalid covariance model')

    # Power spectrum of variable
    syy = np.fft.fftn(np.fft.fftshift(ryy)) / ntot
    syy = np.abs(syy)       # Remove imaginary artifacts
    syy[0] = 0

    real = np.random.randn(*syy.shape)
    imag = np.random.randn(*syy.shape)
    epsilon = real + 1j*imag
    rand = epsilon * np.sqrt(syy)
    Y = np.real(np.fft.ifftn(rand * ntot))

    if mask_given:
        return Y[selection_mask_small]

    return Y



from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext as _build_ext
from setuptools.command.sdist import sdist as _sdist
from os import path



# Numpy stuff
# ----------
# To build the extensions, we need numpy headers. This makes sure numpy is correctly installed.
# See https://stackoverflow.com/questions/27021270/how-to-handle-dependency-on-scipy-in-setup-py
class build_ext(_build_ext):

    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())




# Cython stuff
# ------------
# Here I'm trying to import cython to recreate the *.c files if possible. If not, I'm using the
# existing *.c files for creating the extensions. See here for more info:
# https://stackoverflow.com/questions/4505747/how-should-i-structure-a-python-package-that-contains-cython-code
try:
    from Cython.Build import cythonize
    use_cython = True
    ext = ".pyx"
except ModuleNotFoundError:
    use_cython = False
    ext = ".c"

# Always run cython before creating a source distribution
class sdist(_sdist):
    def run(self):
        # Make sure the compiled Cython files in the distribution are up-to-date
        cythonize(extensions)
        _sdist.run(self)

# Extensions
# ----------
here = path.abspath(path.dirname(__file__))
extensions = [Extension("hyvr.optimized", sources=[path.join(here, "hyvr", "optimized"+ext)]),
              Extension("hyvr.classes.grid", sources=[path.join(here, "hyvr", "classes", "grid"+ext)]),
              Extension("hyvr.classes.contact_surface", sources=[path.join(here, "hyvr", "classes", "contact_surface"+ext)]),
              Extension("hyvr.classes.trough", sources=[path.join(here, "hyvr", "classes", "trough"+ext)]),
              Extension("hyvr.classes.sheet", sources=[path.join(here, "hyvr", "classes", "sheet"+ext)]),
              Extension("hyvr.classes.channel", sources=[path.join(here, "hyvr", "classes", "channel"+ext)]),
              Extension("hyvr.assign_facies", sources=[path.join(here, "hyvr", "assign_facies"+ext)]),
              ]
                         # include_dirs=[np.get_include()])]
if use_cython:
    ext_modules = cythonize(extensions)
else:
    ext_modules = extensions



# Get the long description from the README file
with open(path.join(here, 'README.rst'), mode='r', encoding='utf-8') as f:
    long_description = f.read()

# get version number from file
with open('versionnumber', 'r') as f:
    version = f.read().strip('\n')


# copy the made.ini test case to the package directory
from shutil import copyfile
copyfile(path.join(here, 'testcases', 'made.ini'), path.join(here, 'hyvr', 'made.ini'))

setup(
    name='hyvr',
    version=version,
    description='A python package for simulating hydrogeological virtual realities',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/driftingtides/hyvr',
    author='Jeremy Bennett',
    author_email='hyvr.sim@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Education',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords=['hydrogeology', 'sediment', 'simulator'],
    packages=find_packages(),
    python_requires='>=3.4',
    cmdclass={'build_ext':build_ext,
              'sdist':sdist},
    ext_modules=ext_modules,
    setup_requires=['numpy'],
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'pandas',
        ],
    extras_require = {
        'HDF5': ['h5py'],
        'MODFLOW': ['flopy'],
        'VTR': ['pyevtk'],
        },

    # include testcase config file
    package_data={
        'hyvr': ['made.ini'],
    },
    zip_safe=False, # Due to the .pxd files
)
